/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * This file is part of the libgltf project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#ifndef LIBGLTF_H
#define LIBGLTF_H
#include <stddef.h>
#include "types.h"
#include <GL/glew.h>
#include <glm/glm.hpp>
using namespace glTF;

/** Initialize the glTF renderer - using the content of the .json file
jsonfile: parse this file, and return a glTFHandle which contain the
          informations of files which will be use to draw 3D model by OpenGL.
*/
glTFHandle* gltf_renderer_init(const glTFFile* jsonfile);

/** Set various file contents
handle: It should contain all the buffers of the files and the viewport when
        call this function.
*/
int gltf_renderer_set_content(glTFHandle* handle);

/** Prepare to draw 3D model */
void gltf_render_FPS_enable(glTFHandle* handle);

/** Prepare to draw 3D model */
int gltf_prepare_renderer(glTFHandle* handle);

/** Drawing the 3D model to the current openGLcontext. */
void gltf_renderer(glTFHandle* handle);

/** Complete to draw 3D model */
void gltf_complete_renderer(glTFHandle* handle);

/**Release the data of 3D model.*/
void gltf_renderer_release(glTFHandle* handle);

/*get camera position information*/
void gltf_get_camera_pos(glTFHandle* handle, glm::vec3* pos,
                         glm::vec3* view, glm::vec3* up);

/*get model center position information*/
glm::vec3* gltf_get_model_center_pos(glTFHandle* handle);

/*get camera position information*/
double gltf_get_model_size(glTFHandle* handle);

/** Update camera position, in the time frame of 'time' */
void gltf_renderer_move_camera(glTFHandle* handle, double x, double y,
                               double z, double time);

/** Fly camera position, in the time frame of 'time' */
void gltf_renderer_fly_camera(glTFHandle* handle, glm::vec3 glPosInfo,
                               double time);

/** Update model rotation, to move the mouse in horizontal direction*/
void gltf_renderer_rotate_model(glTFHandle* handle, double horizontal,
                                    double vertical, double planar);

/** Update camera rotation*/
void gltf_renderer_rotate_camera(glTFHandle* handle, double horizontal,
                                 double vertical, double planar);

/** stop model rotation*/
void gltf_renderer_stop_rotate_model(glTFHandle* handle);
/** Enable rotation*/
void gltf_enable_rotation(glTFHandle* handle);

/**Disable rotation*/
void gltf_disable_rotation(glTFHandle* handle);

/** Enable transparency*/
void gltf_enable_transparency(glTFHandle* handle);

/**Disable transparency*/
void gltf_disable_transparency(glTFHandle* handle);

/** Get a bitmap of the screen in the given point in time. */
int gltf_renderer_get_bitmap(glTFHandle** handle, int size,
                              char* buffer, GLenum format, double time = 0);

/** Start Aerial view  */
void gltf_orbit_mode_start(glTFHandle* handle);

/** Stop Aerial view . */
void gltf_orbit_view_stop(glTFHandle* handle);

/** Start playing the glTF animation to the current openGLcontext. */
void gltf_animation_start(glTFHandle* handle);

/** Stop playing the glTF animation. */
void gltf_animation_stop(glTFHandle* handle);

/** Return the state of the glTF animation. */
int gltf_animation_is_playing(glTFHandle* handle);

/** Set the animation to loop. */
void gltf_animation_set_looping(glTFHandle* handle, int loop);

/** Is the animation looping? */
int gltf_animation_get_looping(glTFHandle* handle);

/** Return amount of seconds this glTF animation takes. */
double gltf_animation_get_duration(glTFHandle* handle);

/** Fast forward / rewind to the given point in time. */
void gltf_animation_set_time(glTFHandle* handle, double time);

/** Query where exactly the animation is, in seconds. */
double gltf_animation_get_time(glTFHandle* handle);

void gltf_get_camera_info(glTFHandle* handle, cameraInfo* pCameraInfo);

void gltf_animation_resume(glTFHandle* handle);

//GLdouble* get_glpos_info(glTFHandle* gltfHandle, int x, int y);
#endif

/* vim:set shiftwidth=4 softtabstop=4 expandtab: */
