/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * This file is part of the libgltf project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "Camera.h"
#include <glm/gtx/rotate_vector.hpp>
#include <time.h>
#include <cstring>

#include <math.h>
#define TRACKBALLSIZE  (0.8)
#define AERIALVIEWLIMIT 89.9 * 14

const float PI = float(atan(1.0) * 4.0);

CPhysicalCamera::CPhysicalCamera()
    : vEye(glm::vec3(0.0f, 0.0f, 0.0f))
    , vView(glm::vec3(0.0f, 0.0, -1.0f))
    , vUp(glm::vec3(0.0f, 1.0f, 0.0f))
    , vMove(1.0f)
    , vSpeed(1.0f)
    , fSetTime(0)
    , fTotalMoveTime(0)
    , bRoate_camera_flag(false)
    , bMove_camera_flag(false)
    , bRotateCameraModel(false)
    , bMoveCameraModel(false)
    , bAerialView(false)
    , bAontrolCamera(false)
    , fSensitivity(0.1f)
    , mTrackBallY(1.0f)
    , mTrackBallX(1.0f)
    , mRotateCameraY(1.0f)
    , mRotateCameraX(1.0f)
    , mLook(1.0f)
    , AerialViewY(0)
    , flength(0)
    , vModelCenterPos(1.0f)
    , viewMatrix(1.0f)
    , mProjection(1.0f)
{
    memset(lastquat, 0, sizeof(float) * 4);
    memset(curquatY, 0, sizeof(float) * 4);
    memset(curquatX, 0, sizeof(float) * 4);

    memset(RotateLastquat, 0, sizeof(float) * 4);
    memset(RotateCurquatY, 0, sizeof(float) * 4);
    memset(RotateCurquatX, 0, sizeof(float) * 4);
}

CPhysicalCamera::CPhysicalCamera(glm::vec3 a_vEye,
                                 glm::vec3 a_vView,
                                 glm::vec3 a_vUp,
                                 double a_fSensitivity)
    : vEye(a_vEye)
    , vView(a_vView)
    , vUp(a_vUp)
    , vMove(1.0f)
    , vSpeed(1.0f)
    , fSetTime(0)
    , fTotalMoveTime(0)
    , bRoate_camera_flag(false)
    , bMove_camera_flag(false)
    , bRotateCameraModel(false)
    , bMoveCameraModel(false)
    , bAerialView(false)
    , bAontrolCamera(false)
    , fSensitivity(a_fSensitivity)
    , mTrackBallY(1.0f)
    , mTrackBallX(1.0f)
    , mRotateCameraY(1.0f)
    , mRotateCameraX(1.0f)
    , mLook(1.0f)
    , AerialViewY(0)
    , flength(0)
    , vModelCenterPos(1.0f)
    , viewMatrix(1.0f)
    , mProjection(1.0f)
{
    memset(lastquat, 0, sizeof(float) * 4);
    memset(curquatY, 0, sizeof(float) * 4);
    memset(curquatX, 0, sizeof(float) * 4);

    memset(RotateLastquat, 0, sizeof(float) * 4);
    memset(RotateCurquatY, 0, sizeof(float) * 4);
    memset(RotateCurquatX, 0, sizeof(float) * 4);
}

void CPhysicalCamera::buildRotMatrix(glm::mat4& m, float q[4])
{
    m[0][0] = (float)(1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]));
    m[0][1] = (float)(2.0 * (q[0] * q[1] - q[2] * q[3]));
    m[0][2] = (float)(2.0 * (q[2] * q[0] + q[1] * q[3]));
    m[0][3] = 0.0;

    m[1][0] = (float)(2.0 * (q[0] * q[1] + q[2] * q[3]));
    m[1][1] = (float)(1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]));
    m[1][2] = (float)(2.0 * (q[1] * q[2] - q[0] * q[3]));
    m[1][3] = 0.0;

    m[2][0] = (float)(2.0 * (q[2] * q[0] - q[1] * q[3]));
    m[2][1] = (float)(2.0 * (q[1] * q[2] + q[0] * q[3]));
    m[2][2] = (float)(1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]));
    m[2][3] = 0.0;

    m[3][0] = 0.0;
    m[3][1] = 0.0;
    m[3][2] = 0.0;
    m[3][3] = 1.0;
}

float CPhysicalCamera::GetAngleX()
{
    glm::vec3 vDir = vView - vEye;
    vDir = glm::normalize(vDir);
    glm::vec3 vDir2 = vDir; vDir2.y = 0.0f;
    vDir2 = glm::normalize(vDir2);
    float fAngle = acos(glm::dot(vDir2, vDir)) * (180.0f / PI);
    if (vDir.y < 0)
    {
        fAngle *= -1.0f;
    }
    return fAngle;
}

void CPhysicalCamera::rotateCamera(double horizontal, double vertical,
                                   double /*planar*/)
{
    float deltaX = (float)(horizontal * 0.07);
    float deltaY = (float)(vertical * 0.07);
    glm::vec3 vCameraRotate = vView;
    if (deltaX != 0.0f)
    {
        vCameraRotate -= vEye;
        vCameraRotate = glm::rotate(vCameraRotate, deltaX,
            glm::vec3(0.0f, 1.0f, 0.0f));
        vCameraRotate += vEye;
    }
    if (deltaY != 0.0f)
    {
        glm::vec3 vAxis = glm::cross(vCameraRotate - vEye, vUp);
        vAxis = glm::normalize(vAxis);
        float fAngle = deltaY;
        float fNewAngle = fAngle + GetAngleX();
        if (fNewAngle > -89.80f && fNewAngle < 89.80f)
        {
            vCameraRotate -= vEye;
            vCameraRotate = glm::rotate(vCameraRotate, deltaY, vAxis);
            vCameraRotate += vEye;
        }
    }
    vView = vCameraRotate;
}

void CPhysicalCamera::moveCamera(double x, double y, double z,
                                 double time)
{

    if (time != 0)
    {
        bMove_camera_flag = true;
        fSetTime = time;
        glm::vec3 endView(x, y, z);
        glm::vec3 vRotate = endView - vView;
        vRotate /= time ;
        vSpeed = vRotate;
    }
    else
    {
        if (std::abs(x) > 0.0001 || std::abs(y) > 0.0001 || std::abs(z) > 0.0001)
        {
             bMoveCameraModel = true;
        }
        vMove = glm::vec3(x, y, z);
        fSetTime = 0.0;
    }
}

 void CPhysicalCamera::setLook(glm::mat4 m)
{
    viewMatrix = m;
}

glm::mat4& CPhysicalCamera::getLook()
{
    return viewMatrix;
}

void CPhysicalCamera::setPerspective(glm::mat4 Perspective)
{
    mProjection = Perspective;
}

const glm::mat4 CPhysicalCamera::getPerspective()
{
    return mProjection;
}

glm::mat4 CPhysicalCamera::getModelViewMatrix(glm::mat4 lookat,
                                              glm::mat4& matrix,
                                              float fMoveTime,
                                              float fPreTime)
{
    glm::mat4 modelViewMatrix;
    if (fPreTime <= 0)
    {
        modelViewMatrix = lookat * matrix;
    }
    else
    {
        fTotalMoveTime += fMoveTime;
        if (fTotalMoveTime <= fSetTime)
        {
            if (bRoate_camera_flag == true)
            {
                vView += vSpeed * fMoveTime;
            }
            if (bMove_camera_flag == true)
            {
                vEye += vSpeed * fMoveTime;
                vView += vSpeed * fMoveTime;
            }
            modelViewMatrix = lookat * matrix;
        }
        else
        {
            modelViewMatrix = lookat * matrix;
        }
    }
    return modelViewMatrix;
}

void CPhysicalCamera::RotateObjectMouse(double horizontal, double vertical,
                                        double planar)
{
    if (!bAerialView)
        return;
    if (std::abs(horizontal) > planar || std::abs(vertical) > 0.0001)
    {
        bRotateCameraModel = true;
        float deltaX = (float)(horizontal * fSensitivity * 0.01f);
        float deltaY = (float)(-vertical * fSensitivity * 0.01f);
        AerialViewY += -vertical;
        if (AerialViewY >= -AERIALVIEWLIMIT && AerialViewY <= AERIALVIEWLIMIT)
        {
            trackball(lastquat, 0.0, deltaY, 0.0, 0.0);
            add_quats(lastquat, curquatY, curquatY);
        }
        else
        {
            AerialViewY -= -vertical;
        }
        trackball(lastquat, 0.0f, 0.0f, 0.0f, 0.0f);
        trackball(lastquat, deltaX, 0.0, 0.0, 0.0);
        add_quats(lastquat, curquatX, curquatX);
    }
}

void CPhysicalCamera::removeRotateObject()
{
    trackball(lastquat, 0.0f, 0.0f, 0.0f, 0.0f);
}

void CPhysicalCamera::RotateCamera(double horizontal, double vertical,
                                   double planar)
{
    if (std::abs(horizontal) > planar || std::abs(vertical) > 0.0001)
    {
        bRotateCameraModel = true;
        float deltaX = (float)(horizontal * fSensitivity * 0.01f);
        float deltaY = (float)(-vertical * fSensitivity * 0.01f);
        AerialViewY += -vertical;
        if (AerialViewY >= -AERIALVIEWLIMIT && AerialViewY <= AERIALVIEWLIMIT)
        {
            trackball(RotateLastquat, 0.0, deltaY, 0.0, 0.0);
            add_quats(RotateLastquat, RotateCurquatY, RotateCurquatY);
        }
        else
        {
            AerialViewY -= -vertical;
        }
        trackball(RotateLastquat, 0.0f, 0.0f, 0.0f, 0.0f);
        trackball(RotateLastquat, deltaX, 0.0, 0.0, 0.0);
        add_quats(RotateLastquat, RotateCurquatX, RotateCurquatX);
    }
}

/* vim:set shiftwidth=4 softtabstop=4 expandtab: */
