/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
* This file is part of the libgltf project.
*
* This Source Code Form is subject to the terms of the Mozilla Public
* License, v. 2.0. If a copy of the MPL was not distributed with this
* file, You can obtain one at http://mozilla.org/MPL/2.0/.
*/

#include "Common.h"
#include <cstring>
#include <glm/gtx/transform.hpp>

namespace glTF
{

int gltf_get_file_index_by_name(glTFHandle* handle, const char* name)
{
    if (handle == NULL)
    {
        return LIBGLTF_FILE_NOT_LOAD;
    }
    for (unsigned int i = 0; i < handle->size; i++)
    {
        if ((!strcmp(name, (handle->files[i].filename))))
        {
            return (int)i;
        }
    }
    return LIBGLTF_FILE_NOT_LOAD;
}

/* -- Attribute -- */
Attribute::Attribute()
    : mByteStride(0)
    , mDataType(DataType_UNKNOW)
    , mCount(0)
    , pData(0)
{
}

Attribute::~Attribute()
{
    delete[] pData;
}

void Attribute::setDataCount(unsigned int count)
{
    this->mCount = count;
}

unsigned int Attribute::getDataCount() const
{
    return this->mCount;
}

void Attribute::setByteStride(unsigned int byteStride)
{
    this->mByteStride = byteStride;
}

unsigned int Attribute::getByteStride() const
{
    return this->mByteStride;
}

void Attribute::setDataType(DataType_E type)
{
    this->mDataType = type;
}

DataType_E Attribute::getDataType() const
{
    return this->mDataType;
}

void Attribute::setAttributeData(const char* pValue, unsigned int length)
{
    if (0 == pData)
    {
        pData = new char[length];
    }
    memcpy(this->pData, pValue, length);
}

const char* Attribute::getAttributeData() const
{
    return pData;
}

/* -- Light -- */
Light::Light()
    : mName("light0Color")
    , mType(LightSource_UNDEFINED)
    , mAttenuationConstant(0)
    , mAttenuationLinear(0)
    , mAttenuationQuadratic(0)
    , mColor(1.0f, 1.0f, 1.0f)
{

}

Light::Light(const Light& light)
    : mName(light.mName)
    , mType(light.mType)
    , mAttenuationConstant(light.mAttenuationConstant)
    , mAttenuationLinear(light.mAttenuationLinear)
    , mAttenuationQuadratic(light.mAttenuationQuadratic)
    , mColor(light.mColor)
{
}

Light::~Light()
{

}

void Light::setLightName(std::string strName)
{
    this->mName = strName;
}

const std::string Light::getLightName()
{
    return mName;
}

void Light::setType(LightSourceType_E Type)
{
    this->mType = Type;
}

void Light::setColor(glm::vec3 Color)
{
    this->mColor = Color;
}

const glm::vec3 Light::getColor()
{
    return this->mColor;
}

void Light::setAttenuationConstant(float AttenuationConstant)
{
    this->mAttenuationConstant = AttenuationConstant;
}

float Light::getAttenuationConstant()
{
    return mAttenuationConstant;
}

void Light::setAttenuationLinear(float AttenuationLinear)
{
    this->mAttenuationLinear = AttenuationLinear;
}

float Light::getmAttenuationLinear()
{
    return mAttenuationLinear;
}

void Light::setAttenuationQuadratic(float AttenuationQuadratic)
{
    this->mAttenuationQuadratic = AttenuationQuadratic;
}

float Light::getAttenuationQuadratic()
{
    return mAttenuationQuadratic;
}

/* -- Camera -- */
ParseCamera::ParseCamera()
    : mName()
    , mXFov(37.8492f)
    , mYFov(1.0f)
    , mNear(1.0f)
    , mFar(500000.0f)
    , mAspectRatio(1.5f)
    , cameraNode(0)
{
}

ParseCamera::~ParseCamera()
{
}

void ParseCamera::setXFov(float xFov)
{
    this->mXFov = xFov;
}

float ParseCamera::getXFov()
{
    return this->mXFov;
}

void ParseCamera::setYFov(float yFov)
{
    this->mYFov = yFov;
}

float ParseCamera::getYFov()
{
    return this->mYFov;
}

void ParseCamera::setFar(float far)
{
    this->mFar = far;
}

float ParseCamera::getFar()
{
    return this->mFar;
}

void ParseCamera::setNear(float near)
{
    this->mNear = near;
}

float ParseCamera::getNear()
{
    return this->mNear;
}

void ParseCamera::setAspectRatio(float aspectRatio)
{
    this->mAspectRatio = aspectRatio;
}

float ParseCamera::getAspectRatio()
{
    return this->mAspectRatio;
}

void ParseCamera::setCameraNode(Node* pNode)
{
    this->cameraNode = pNode;
}

Node* ParseCamera::getCameraNode()
{
    return this->cameraNode;
}

/* -- TechAttribute -- */
TechAttribute::TechAttribute()
    : mAttributeName()
    , mAttributeIndex()
{
}

TechAttribute::~TechAttribute()
{
}

void TechAttribute::setAttributeName(const std::string& name)
{
    this->mAttributeName = name;
}

const std::string TechAttribute::getAttributeName()
{
    return this->mAttributeName;
}

void TechAttribute::setAttributeIndex(const std::string& index)
{
    this->mAttributeIndex = index;
}

const std::string TechAttribute::getAttributeIndex()
{
    return this->mAttributeIndex;
}

/* -- TechUniform -- */
TechUniform::TechUniform()
    : mUniformName()
    , mUniformIndex()
{
}

TechUniform::~TechUniform()
{
}

void TechUniform::setUniformName(const std::string& name)
{
    this->mUniformName = name;
}

void TechUniform::setUniformIndex(const std::string& index)
{
    this->mUniformIndex = index;
}

/* -- Technique -- */
Technique::Technique()
    : mShaderProg()
    , mTechAttrMap()
    , mTechUnifVec()
    , mTLight()
    , mProgramId(0)
    , mTechniqueId()
    , mVShaderName()
    , mFShaderName()
    , mTechniqueState(false)
    , pState(new TechniqueState())
{
}

Technique::~Technique()
{
    std::vector<TechUniform*>::iterator vit;
    for (vit = mTechUnifVec.begin(); vit != mTechUnifVec.end(); ++vit)
    {
        delete(*vit);
    }

    std::map<std::string, TechAttribute*>::iterator mit;
    for (mit = mTechAttrMap.begin(); mit != mTechAttrMap.end(); ++mit)
    {
        delete mit->second;
    }
    mTechAttrMap.clear();

    std::vector<techLight*>::iterator vlit;
    for (vlit = mTLight.begin(); vlit != mTLight.end(); ++vlit)
    {
        delete(*vlit);
    }
    mTLight.clear();

    if (this->mTechniqueState)
        freeTechnique();
    delete pState;
}

void Technique::setProgramState(bool flag)
{
    this->mTechniqueState = flag;
}

bool Technique::getProgramState() const
{
    return this->mTechniqueState;
}

unsigned int Technique::getProgramId() const
{
    return this->mProgramId;
}

void Technique::setTechId(std::string techId)
{
    this->mTechniqueId = techId;
}

const std::string& Technique::getTechId()
{
    return this->mTechniqueId;
}

void Technique::setVertexShader(const std::string& shaderName)
{
    this->mVShaderName = shaderName;
}

void Technique::setFragmentShader(const std::string& shaderName)
{
    this->mFShaderName = shaderName;
}

void Technique::insertTechAttribute(const std::string& key,
                                    TechAttribute* pTechAttr)
{
    this->mTechAttrMap.insert(
        std::map<std::string, TechAttribute*>::value_type(key, pTechAttr));
}

TechAttribute* Technique::getTechAttribute(const std::string& key)
{
    std::map<std::string, TechAttribute*>::const_iterator it;
    it = mTechAttrMap.find(key);
    if (it != mTechAttrMap.end())
    {
        return it->second;
    }
    return 0;
}

void Technique::pushTechUniform(TechUniform* pUniform)
{
    (this->mTechUnifVec).push_back(pUniform);
}

void Technique::pushTLight(techLight* ptLight)
{
    (this->mTLight).push_back(ptLight);
}

const std::vector<techLight*> Technique::poptLight()
{
    return mTLight;
}

TechniqueState* Technique::getTechState()
{
    return pState;
}

int Technique::initTechnique(glTFHandle* pGltfHandle)
{
    int vShaderIdx = gltf_get_file_index_by_name(pGltfHandle,
                                                 mVShaderName.c_str());
    int fShaderIdx = gltf_get_file_index_by_name(pGltfHandle,
                                                 mFShaderName.c_str());
    if (vShaderIdx >= 0 && fShaderIdx >= 0)
    {
        const char* pvShader = pGltfHandle->files[vShaderIdx].buffer;
        size_t ivShaderSize = pGltfHandle->files[vShaderIdx].size;
        const char* pfShader = pGltfHandle->files[fShaderIdx].buffer;
        size_t ifShaderSize = pGltfHandle->files[fShaderIdx].size;
        mProgramId = mShaderProg.createProgram(pvShader, ivShaderSize,
                                               pfShader, ifShaderSize);
    }

    if (0 != mProgramId)
    {
        this->setProgramState(true);
        return LIBGLTF_SUCCESS;
    }
    else
    {
        return LIBGLTF_SHADER_ERROR;
    }
}

bool Technique::useTechnique()
{
    if (this->mTechniqueState)
    {
        mShaderProg.useProgram(0);
        mShaderProg.useProgram(mProgramId);
    }
    return this->mTechniqueState;
}

void Technique::freeTechnique()
{
    mShaderProg.deleteProgram(mProgramId);
    this->setProgramState(false);
}

/* -- MaterialProperty -- */
MaterialProperty::MaterialProperty()
    : mPropertyName()
    , mTextureType(TextureType_NONE)
    , mImagePath()
    , mDataType(DataType_UNKNOW)
    , mDataLength(0)
    , pData(0)
{
}

MaterialProperty::~MaterialProperty()
{
    delete [] pData;
}

void MaterialProperty::setPropertyName(const std::string& name)
{
    this->mPropertyName = name;
}

const std::string& MaterialProperty::getPropertyName() const
{
    return this->mPropertyName;
}

void MaterialProperty::setImagePath(const std::string& index)
{
    this->mImagePath = index;
}

const std::string& MaterialProperty::getImagePath() const
{
    return this->mImagePath;
}

void MaterialProperty::setDataLength(unsigned int length)
{
    this->mDataLength = length;
}

void MaterialProperty::setDataType(DataType_E type)
{
    this->mDataType = type;
}

DataType_E MaterialProperty::getDataType() const
{
    return this->mDataType;
}

void MaterialProperty::setPropertyData(char* pValue, unsigned int size)
{
    if (0 != this->pData)
        delete [] this->pData;
    this->pData = new char[size];
    memcpy(this->pData, pValue, size);
}

const char* MaterialProperty::getPropertyData() const
{
    return pData;
}

/* -- Material -- */
Material::Material()
    : mTechniqueId()
    , mPropertyVec()
{
}

Material::~Material()
{
    std::vector<MaterialProperty*>::iterator it;
    for (it = mPropertyVec.begin(); it != mPropertyVec.end(); ++it)
    {
        delete(*it);
    }
    mPropertyVec.clear();
}

void Material::pushMaterialProper(MaterialProperty* pProperty)
{
    this->mPropertyVec.push_back(pProperty);
}

const MaterialProperty* Material::getMaterialProper(unsigned int index)
{
    if (this->mPropertyVec.size() <= index)
    {
        return 0;
    }
    return this->mPropertyVec[index];
}

unsigned int Material::getMaterialProperSize()
{
    return this->mPropertyVec.size();
}

void Material::setTechniqueId(const std::string& techniqueId)
{
    this->mTechniqueId = techniqueId;
}

const std::string& Material::getTechniqueId() const
{
    return this->mTechniqueId;
}

/* -- Primitives -- */
Primitives::Primitives()
    : mAttributeMap()
    , mIndicesIndex()
    , mMaterialIndex()
{
}

Primitives::~Primitives()
{
    mAttributeMap.clear();
}

void Primitives::setMaterialIndex(std::string materialIndex)
{
    this->mMaterialIndex = materialIndex;
}

const std::string Primitives::getMaterialIndex() const
{
    return this->mMaterialIndex;
}

void Primitives::setIndicesIndex(std::string indicesIndex)
{
    this->mIndicesIndex = indicesIndex;
}

const std::string Primitives::getIndicesIndex() const
{
    return this->mIndicesIndex;
}

void Primitives::insertAttribute(const std::string& key,
                                 const std::string& value)
{
    mAttributeMap.insert(
        std::map<std::string, std::string>::value_type(key, value));
}

const std::string Primitives::getAttributeIndex(const std::string& key) const
{
    std::map<std::string, std::string>::const_iterator it;
    it = mAttributeMap.find(key);
    if (it != mAttributeMap.end())
    {
        return it->second;
    }

    return "";
}

/* -- Mesh -- */
Mesh::Mesh()
    : mName()
    , mPrimitiveVec()
{
}

Mesh::~Mesh()
{
    std::vector<Primitives*>::iterator it;
    for (it = mPrimitiveVec.begin(); it != mPrimitiveVec.end(); ++it)
    {
        delete(*it);
    }
    mPrimitiveVec.clear();
}

void Mesh::setMeshName(const std::string& mName)
{
    this->mName = mName;
}

void Mesh::setPrimitiveVec(Primitives* pPrim)
{
    (this->mPrimitiveVec).push_back(pPrim);
}

const Primitives* Mesh::getPrimitiveVec(unsigned int index) const
{
    if (this->mPrimitiveVec.size() <= index)
    {
        return 0;
    }
    return this->mPrimitiveVec[index];
}

unsigned int Mesh::getPrimitiveVecSize() const
{
    return this->mPrimitiveVec.size();
}

/* -- Node -- */
Node::Node()
    : mNodeType(NodeType_Init)
    , mName()
    , mLocalMatrix(1.0)
    , mGlobalMatrix(1.0)
    , mScaleMatrix(1.0)
    , mTransMatrix(1.0)
    , mRotateMatrix(1.0)
    , pParent(0)
    , mChildrenVec()
    , mMesheIndexVec()
    , mCameraIndex()
    , mLightIndex()
    , mJointIndex()
    , mSkinIndex()
    , pSkin(0)
    , mBoneVec()
    , pAnimation(0)
    , mSkeleIndex()
    , bJointFlag(false)
    , bMatrixFlag(true)
    , bUpdateFlag(false)
{
}

Node::~Node()
{
    std::vector<Node*>::iterator it;
    for (it = mChildrenVec.begin(); it != mChildrenVec.end(); ++it)
    {
        delete(*it);
    }
    mChildrenVec.clear();

    mBoneVec.clear();
}

void Node::setGlobalMatrix(const glm::mat4& matrix)
{
    this->mGlobalMatrix = matrix;
}

glm::mat4& Node::getGlobalMatrix()
{
    return this->mGlobalMatrix;
}

void Node::setLocalMatrix(const glm::mat4& matrix)
{
    this->mLocalMatrix = matrix;
}

glm::mat4& Node::getLocalMatrix()
{
    return this->mLocalMatrix;
}

void Node::setRotate(const glm::mat4& rotate)
{
    this->mRotateMatrix = rotate;
}

void Node::setRotate(const float* pbuf)
{
    glm::vec4 rotate;
    memcpy(&rotate, pbuf, sizeof(glm::vec4));
    this->mRotateMatrix = glm::rotate(rotate.w*57, glm::vec3(rotate));
}

const glm::mat4& Node::getRotate()
{
    return this->mRotateMatrix;
}

void Node::setScale(const glm::mat4& scale)
{
    this->mScaleMatrix = scale;
}

void Node::setScale(const float* pbuf)
{
    glm::vec3 scale;
    memcpy(&scale, pbuf, sizeof(glm::vec3));
    this->mScaleMatrix = glm::scale(scale);
}

const glm::mat4& Node::getScale()
{
    return this->mScaleMatrix;
}

void Node::setTranslate(const glm::mat4& trans)
{
    this->mTransMatrix = trans;
}

void Node::setTranslate(const float* pbuf)
{
    glm::vec3 trans;
    memcpy(&trans, pbuf, sizeof(glm::vec3));
    this->mTransMatrix = glm::translate(trans);
}

const glm::mat4& Node::getTranslate()
{
    return this->mTransMatrix;
}

void Node::setNodeType(NodeType_E _type)
{
    this->mNodeType = _type;
}

void Node::setNodeName(const std::string& name)
{
    this->mName = name;
}

const std::string& Node::getNodeName() const
{
    return this->mName;
}

void Node::setJointId(const std::string& jointId)
{
    this->mJointIndex = jointId;
}

const std::string& Node::getJointId() const
{
    return this->mJointIndex;
}

void Node::setLightIndex(const std::string& lightIndex)
{
    this->mLightIndex = lightIndex;
}

const std::string Node::getLightIndex() const
{
    return this->mLightIndex;
}

void Node::setSkinIndex(const std::string& index)
{
    this->mSkinIndex = index;
}

const std::string& Node::getSkinIndex() const
{
    return this->mSkinIndex;
}

void Node::setSkinPoint(Skin* pSkin)
{
    this->pSkin = pSkin;
}

Skin* Node::getSkinPoint()
{
    return this->pSkin;
}

void Node::setAnimPoint(Animation* pAnima)
{
    this->pAnimation = pAnima;
}

Animation* Node::getAnimPoint()
{
    return this->pAnimation;
}

void Node::setSkeleIndex(const std::string& index)
{
    this->mSkeleIndex = index;
}

const std::string& Node::getSkeleIndex() const
{
    return this->mSkeleIndex;
}

void Node::pushBoneNode(Node* pBone)
{
    (this->mBoneVec).push_back(pBone);
}

Node* Node::getBoneNode(unsigned int index)
{
    if (this->mBoneVec.size() <= index)
    {
        return 0;
    }
    return this->mBoneVec[index];
}

unsigned int Node::getBoneNodeSize()
{
    return (this->mBoneVec).size();
}

void Node::setCameraIndex(std::string cameraIndex)
{
    this->mCameraIndex = cameraIndex;
}

const std::string Node::getCameraIndex()
{
    return this->mCameraIndex;
}

void Node::pushMeshIndex(const std::string& _index)
{
    (this->mMesheIndexVec).push_back(_index);
}

const std::string Node::getMeshIndex(unsigned int index) const
{
    if (this->mMesheIndexVec.size() <= index)
    {
        return 0;
    }
    return this->mMesheIndexVec[index];
}

unsigned int Node::getMeshIndexSize() const
{
    return this->mMesheIndexVec.size();
}

void Node::pushChildNode(Node* pNode)
{
    (this->mChildrenVec).push_back(pNode);
}

Node* Node::getChildNode(unsigned int index)
{
    if (this->mChildrenVec.size() <= index)
    {
        return 0;
    }
    return this->mChildrenVec[index];
}

unsigned int Node::getChildNodeSize() const
{
    return this->mChildrenVec.size();
}

void Node::setParentNode(Node* pParentNode)
{
    this->pParent = pParentNode;
}

Node* Node::getParentNode()
{
    return this->pParent;
}

void Node::setJointFlag(bool flag)
{
    this->bJointFlag = flag;
}

bool Node::getJointFlag()
{
    return this->bJointFlag;
}

void Node::setMatrixFlag(bool flag)
{
    this->bMatrixFlag = flag;
}

bool Node::getMatrixFlag()
{
    return this->bMatrixFlag;
}

void Node::setUpdateFlag(bool flag)
{
    this->bUpdateFlag = flag;
}

bool Node::getUpdateFlag()
{
    return this->bUpdateFlag;
}



/* -- Animation -- */
Animation::Animation()
    : mBoneId()
    , mDuration(0)
    , mChannelBits(0)
    , mTimeValueVec()
{
}

Animation::~Animation()
{
    mTimeValueVec.clear();
}

void Animation::setDuration(double duration)
{
    this->mDuration = duration;
}

double Animation::getDuration()
{
    return this->mDuration;
}

const glm::mat4& Animation::findTimeValue(double time)
{
    std::vector<QuatKey>::const_iterator locIt;
    locIt = std::lower_bound(mTimeValueVec.begin(), mTimeValueVec.end(),
                             time, Animation::compareQuatKey);
    return locIt->mValue;
}

void Animation::pushTimeValue(double time, const glm::mat4& matrix)
{
    QuatKey quatKey;
    quatKey.mTime = time;
    quatKey.mValue = matrix;
    this->mTimeValueVec.push_back(quatKey);
}

void Animation::setTimeValue(const glm::mat4& matrix, unsigned int index)
{
    this->mTimeValueVec[index].mValue = matrix;
}

const glm::mat4& Animation::getTimeValue(unsigned int index)
{
    return this->mTimeValueVec[index].mValue;
}

void Animation::setBoneId(std::string boneId)
{
    this->mBoneId = boneId;
}

const std::string& Animation::getBoneId()
{
    return this->mBoneId;
}

/* -- Skin -- */
Skin::Skin()
    : mName()
    , pBindMatrix(0)
    , mBindMatrixCount(0)
    , mBoneIdVec()
{
}

Skin::~Skin()
{
    delete[] pBindMatrix;
    mBoneIdVec.clear();
}

void Skin::setSkinName(const std::string& name)
{
    this->mName = name;
}

const std::string& Skin::getSkinName()
{
    return this->mName;
}

void Skin::setBindMatrix(glm::mat4* matrix)
{
    this->pBindMatrix = matrix;
}

glm::mat4* Skin::getBindMatrix()
{
    return this->pBindMatrix;
}

void Skin::setBindMatrixCount(unsigned int count)
{
    this->mBindMatrixCount = count;
}

unsigned int Skin::getBindMatrixCount()
{
    return this->mBindMatrixCount;
}

void Skin::pushBoneId(std::string boneId)
{
    this->mBoneIdVec.push_back(boneId);
}

const std::string Skin::getBoneId(unsigned int index)
{
    if (this->mBoneIdVec.size() <= index)
    {
        return 0;
    }
    return this->mBoneIdVec[index];
}

unsigned int Skin::getBoneIdSize()
{
    return this->mBoneIdVec.size();
}

/* -- Scene -- */
Scene::Scene()
    : mAnimaMap()
    , mDuration(0.0)
    , mLightNodeMap()
    , mSkinVec()
    , pRootNode(0)
    , mNodeVec()
    , mMeshMap()
    , mMaterialMap()
    , mLightMap()
    , mCameraMap()
    , mVertexMax(-1e10f, -1e10f, -1e10f)
    , mVertexMin(1e10f, 1e10f, 1e10f)
    , mAttributeMap()
    , mTechniqueVec()
    , pBuffer(0)
    , bUseCameraInJson(false)
    , pGltfHandle(0)
    , mTexturesMap()
{
}

Scene::~Scene()
{
    freeMap(mMeshMap);
    freeMap(mMaterialMap);
    freeMap(mLightMap);
    freeMap(mCameraMap);
    //freeMap(mAttributeMap);
    freeMap(mTexturesMap);
    freeMap(mAnimaMap);
    //delete[] pBuffer;
    delete pRootNode;
    mNodeVec.clear();
    mLightNodeMap.clear();

    std::vector<Technique*>::iterator it;
    for (it = mTechniqueVec.begin(); it != mTechniqueVec.end(); ++it)
    {
        delete *it;
    }
    mTechniqueVec.clear();

    std::vector<Skin*>::iterator itSkin;
    for (itSkin = mSkinVec.begin(); itSkin != mSkinVec.end(); ++itSkin)
    {
        delete *itSkin;
    }
    mSkinVec.clear();
}

template<typename M> void Scene::freeMap(M& aMap)
{
    for (typename M::iterator it = aMap.begin(); it != aMap.end(); ++it)
    {
        delete it->second;
    }
    aMap.clear();
}

Animation* Scene::findAnimation(const std::string& nodeId)
{
    std::map<std::string, Animation*>::iterator it;
    it = mAnimaMap.find(nodeId);
    if (it != mAnimaMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertAnimMap(const std::string& key, Animation* pAnim)
{
    this->mAnimaMap.insert(
        std::map<std::string, Animation*>::value_type(key, pAnim));
}

unsigned int Scene::getAnimationCount()
{
    return this->mAnimaMap.size();
}

void Scene::insertLightNodeMap(const std::string& LightIndex, Node* pNode)
{
    this->mLightNodeMap.insert(
        std::map<std::string, Node*>::value_type(LightIndex, pNode));
}

Node* Scene::findLightNodeMap(std::string key)
{
    std::map<std::string, Node*>::iterator it;
    it = mLightNodeMap.find(key);
    if (it != mLightNodeMap.end())
    {
        return it->second;
    }
    return NULL;
}

Mesh* Scene::findMesh(const std::string& key)
{
    std::map<std::string, Mesh*>::iterator it;
    it = mMeshMap.find(key);
    if (it != mMeshMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertMeshMap(const std::string& key, Mesh* pMesh)
{
    this->mMeshMap.insert(
        std::map<std::string, Mesh*>::value_type(key, pMesh));
}

Material* Scene::findMaterial(const std::string& key)
{
    std::map<std::string, Material*>::iterator it;
    it = mMaterialMap.find(key);
    if (it != mMaterialMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertMaterialMap(const std::string& key, Material* pMaterial)
{
    this->mMaterialMap.insert(
        std::map<std::string, Material*>::value_type(key, pMaterial));
}

ParseCamera* Scene::findCamera(const std::string& key)
{
    std::map<std::string, ParseCamera*>::iterator it;
    it = mCameraMap.find(key);
    if (it != mCameraMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertCameraMap(const std::string& key, ParseCamera* pCamera)
{
    this->mCameraMap.insert(
        std::map<std::string, ParseCamera*>::value_type(key, pCamera));
}

std::string Scene::getCameraIndex()
{
    return "";
}

Light* Scene::findLight(const std::string& key)
{
    std::map<std::string, Light*>::iterator it;
    it = mLightMap.find(key);
    if (it != mLightMap.end())
    {
        return it->second;
    }
    return 0;
}

const Attribute* Scene::findAttribute(const std::string& key)
{
    std::map<std::string, Attribute*>::iterator it;
    it = mAttributeMap.find(key);
    if (it != mAttributeMap.end())
    {
        return it->second;
    }
    return 0;
}

void Scene::insertAttributeMap(std::string key, Attribute* pAttribute)
{
    this->mAttributeMap.insert(
        std::map<std::string, Attribute*>::value_type(key, pAttribute));
}

void Scene::clearAttributeMap()
{
    freeMap(mAttributeMap);
}

void Scene::insertLightMap(const std::string& key, Light* pLight)
{
    this->mLightMap.insert(
        std::map<std::string, Light*>::value_type(key, pLight));
}

Node* Scene::getRootNode()
{
    return this->pRootNode;
}

void Scene::setRootNode(Node* pNode)
{
    this->pRootNode = pNode;
}

void Scene::pushNode(Node* pNode)
{
    this->mNodeVec.push_back(pNode);
}

Node* Scene::getNode(unsigned int index)
{
    if (this->mNodeVec.size() <= index)
    {
        return 0;
    }
    return this->mNodeVec[index];
}

unsigned int Scene::getNodeSize()
{
    return this->mNodeVec.size();
}

void Scene::pushSkin(Skin* pSkin)
{
    this->mSkinVec.push_back(pSkin);
}

Skin* Scene::getSkin(unsigned int index)
{
    if (this->mSkinVec.size() <= index)
    {
        return 0;
    }
    return this->mSkinVec[index];
}

unsigned int Scene::getSkinSize()
{
    return this->mSkinVec.size();
}

char* Scene::getBuffer(unsigned int length)
{
    if (0 == this->pBuffer && 0 != length)
    {
        this->pBuffer = new char[length];
    }
    return this->pBuffer;
}

int Scene::setBuffer(const std::string& binName, unsigned int length)
{
    int idx = gltf_get_file_index_by_name(pGltfHandle, binName.c_str());
    if (idx != LIBGLTF_FILE_NOT_LOAD)
    {
        if (pGltfHandle->files[idx].buffer != NULL)
        {
            memcpy(pBuffer, pGltfHandle->files[idx].buffer, length);
            return idx;
        }
    }
    return idx;
}

void Scene::freeBuffer()
{
    if (pBuffer != 0)
    {
        delete [] pBuffer;
        pBuffer = 0;
    }
}

void Scene::setVertexMax(float x, float y, float z)
{
    mVertexMax.x = std::max(x, mVertexMax.x);
    mVertexMax.y = std::max(y, mVertexMax.y);
    mVertexMax.z = std::max(z, mVertexMax.z);
}

glm::vec3& Scene::getVertexMax()
{
    return mVertexMax;
}

void Scene::setVertexMin(float x, float y, float z)
{
    mVertexMin.x = std::min(x, mVertexMin.x);
    mVertexMin.y = std::min(y, mVertexMin.y);
    mVertexMin.z = std::min(z, mVertexMin.z);
}

glm::vec3& Scene::getVertexMin()
{
    return mVertexMin;
}

void Scene::pushTechnique(class Technique* pTechnique)
{
    this->mTechniqueVec.push_back(pTechnique);
}

class Technique* Scene::getTechnique(unsigned int index)
{
    if (this->mTechniqueVec.size() <= index)
    {
        return 0;
    }
    return this->mTechniqueVec[index];
}

unsigned int Scene::getTechSize()
{
    return this->mTechniqueVec.size();
}

bool Scene::loadTexture(const std::string& imagePath)
{
    if (findTexture(imagePath) == NULL)
    {
        Texture* oneTexture = new Texture();

        glTFFile* pglTFFile = getGltfFileByFileName(imagePath);
        oneTexture->loadTexture2DFromBuffer((unsigned char*)
                                            (pglTFFile->buffer),
                                            pglTFFile->imagewidth,
                                            pglTFFile->imageheight);
        oneTexture->setFiltering(TEXTURE_FILTER_MAG_NEAREST,
                                 TEXTURE_FILTER_MIN_NEAREST);
        insertTextureMap(imagePath, oneTexture);
    }
    return true;
}

Texture* Scene::findTexture(const std::string& key)
{
    std::map<std::string, Texture*>::iterator it;
    it = mTexturesMap.find(key);
    if (it != mTexturesMap.end())
    {
        return it->second;
    }
    return NULL;
}

void Scene::insertTextureMap(const std::string& key, Texture* pTexture)
{
    this->mTexturesMap.insert(
        std::map<std::string, Texture*>::value_type(key, pTexture));
}

double Scene::getDuration()
{
    return this->mDuration;
}

void Scene::setDuration(double duration)
{
    this->mDuration = duration;
}

void Scene::setGltfHandle(glTFHandle* handle)
{
    if (handle != NULL)
    {
        this->pGltfHandle = handle;
    }
}

glTFHandle* Scene::getGltfHandle()
{
    return this->pGltfHandle;
}

glTFFile* Scene::getGltfFileByFileName(const std::string& fileName)
{
    if (pGltfHandle == NULL)
    {
        return NULL;
    }
    for (unsigned int i = 0; i < pGltfHandle->size; i++)
    {
        if (!strcmp(pGltfHandle->files[i].filename, fileName.c_str()))
        {
            return &pGltfHandle->files[i];
        }
    }
    return NULL;
}
}

/* vim:set shiftwidth=4 softtabstop=4 expandtab: */
