/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * This file is part of the libgltf project.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#include "RenderScene.h"
#include "GLTFTime.h"
#include <glm/gtx/transform.hpp>
#include <glm/gtc/matrix_inverse.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>
#include <cerrno>


RenderWithFBO::RenderWithFBO()
    : mFboProId(0)
    , mFboId(0)
    , mRboId(0)
    , mTexId(0)
    , mShotFboId(0)
    , mShotRboId(0)
    , mShotTexId(0)
    , mInverseFboId(0)
    , mInverseRboId(0)
    , mInverseTexId(0)
    , mVertexBuf(0)
    , mTexCoordBuf(0)
{
}

void RenderWithFBO::inverseBitMap(int width, int height)
{
    GLuint proId = loadFboShader(INVERSEVERT, INVERSEFRAG);
    GLuint texCoordBuf = 0;
    GLuint vertexBuf = 0;
    GLfloat coordVertices[] =
    {
        0.0f, 1.0f,
        1.0f, 1.0f,
        1.0f, 0.0f,
        0.0f, 0.0f,
    };
    GLfloat squareVertices[] =
    {
        -1.0f, -1.0f, -1.0,
         1.0f, -1.0f, -1.0,
         1.0f,  1.0f, -1.0,
        -1.0f,  1.0f, -1.0
    };
    setBufferForFbo(texCoordBuf, vertexBuf, coordVertices,
                    sizeof(coordVertices), squareVertices,
                    sizeof(squareVertices));
    createAndBindInverseFBO(width, height);
    glViewport(0, 0, width, height);
    inverseTexture(proId, texCoordBuf, vertexBuf);
}

void RenderWithFBO::releaseBitMapFBO()
{
    glDeleteFramebuffers(1, &mInverseFboId);
    glDeleteRenderbuffers(1, &mInverseRboId);
    glDeleteTextures(1, &mInverseTexId);

}

void RenderWithFBO::releaseInverseBitMapFBO()
{
    glDeleteFramebuffers(1, &mShotFboId);
    glDeleteRenderbuffers(1, &mShotRboId);
    glDeleteTextures(1, &mShotTexId);
}

GLuint RenderWithFBO::loadFboShader(const char* vShader,
                                    const char* fShader)
{
    ShaderProgram shaderPro;
    GLuint progId = glCreateProgram();
    if(false == shaderPro.loadShader(progId, vShader,
                                     strlen(vShader), GL_VERTEX_SHADER))
    {
        return 0;
    }
    if(false == shaderPro.loadShader(progId, fShader,
                                     strlen(fShader), GL_FRAGMENT_SHADER))
    {
        return 0;
    }
    return progId;
}

void RenderWithFBO::setBufferForFbo(GLuint& texCoordBuf, GLuint& vertexBuf,
                                    GLfloat* pCoord, GLuint numCoord,
                                    GLfloat* pSquare, GLuint numSquare)
{
    glGenBuffers(1, &texCoordBuf);
    glBindBuffer(GL_ARRAY_BUFFER, texCoordBuf);
    glBufferData(GL_ARRAY_BUFFER, numCoord, pCoord, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glGenBuffers(1, &vertexBuf);
    glBindBuffer(GL_ARRAY_BUFFER, vertexBuf);
    glBufferData(GL_ARRAY_BUFFER, numSquare, pSquare, GL_STATIC_DRAW);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void RenderWithFBO::createAndBindInverseFBO(int width, int height)
{
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    createRenderObj(width, height, mInverseRboId);
    createTextureObj(width, height, mInverseTexId);
    createFrameBufferObj(mInverseFboId, mInverseTexId, mInverseRboId);
    glBindFramebuffer(GL_FRAMEBUFFER, mInverseFboId);
}

void RenderWithFBO::createAndBindBitMapFBO(int width, int height)
{
    createRenderObj(width, height, mShotRboId);
    createTextureObj(width, height, mShotTexId);
    createFrameBufferObj(mShotFboId, mShotTexId, mShotRboId);
    glBindFramebuffer(GL_FRAMEBUFFER, mShotFboId);
}

void RenderWithFBO::inverseTexture(GLuint proId, GLuint texCoordBuf,
                                   GLuint vertexBuf)
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    GLuint texCoordId = glGetAttribLocation(proId, "texCoord");
    GLuint vertexId   = glGetAttribLocation(proId, "vPosition");
    GLuint textureId  = glGetUniformLocation(proId, "RenderTex");

    glUseProgram(proId);
    glEnableVertexAttribArray(vertexId);
    glBindBuffer(GL_ARRAY_BUFFER, vertexBuf);
    glVertexAttribPointer(vertexId, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    glEnableVertexAttribArray(texCoordId);
    glBindBuffer(GL_ARRAY_BUFFER, texCoordBuf);
    glVertexAttribPointer(texCoordId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, mShotTexId);
    glUniform1i(textureId, 0);
    glDrawArrays(GL_QUADS, 0, 4);
    glDisableVertexAttribArray(vertexId);
    glDisableVertexAttribArray(texCoordId);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
    glUseProgram(0);
}

void RenderWithFBO::createRenderObj(int width, int height, GLuint& rboId)
{
    glGenRenderbuffers(1, &rboId);
    glBindRenderbuffer(GL_RENDERBUFFER, rboId);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
}

void RenderWithFBO::createTextureObj(int width, int height, GLuint& texId)
{
    glGenTextures(1, &texId);
    glBindTexture(GL_TEXTURE_2D, texId);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB,
                 GL_UNSIGNED_BYTE, NULL);
    glBindTexture(GL_TEXTURE_2D, 0);
}

int RenderWithFBO::createFrameBufferObj(GLuint& fboId, GLuint texId,
                                        GLuint rboId)
{
    int status;
    glGenFramebuffers(1, &fboId);
    status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (GL_FRAMEBUFFER_COMPLETE != status)
    {
        return LIBGLTF_CREATE_FBO_ERROR;
    }
    glBindFramebuffer(GL_FRAMEBUFFER, fboId);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, texId);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
        GL_TEXTURE_2D, texId, 0);
    status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (GL_FRAMEBUFFER_COMPLETE != status)
    {
        return LIBGLTF_CREATE_FBO_ERROR;
    }
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);

    glBindRenderbuffer(GL_RENDERBUFFER, rboId);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
                              GL_RENDERBUFFER, rboId);
    status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (GL_FRAMEBUFFER_COMPLETE != status)
    {
        return LIBGLTF_CREATE_FBO_ERROR;
    }
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    return LIBGLTF_SUCCESS;
}

int RenderWithFBO::renderFboTexture()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glUseProgram(mFboProId);

    GLuint texCoordId = glGetAttribLocation(mFboProId, "texCoord");
    GLuint vertexId   = glGetAttribLocation(mFboProId, "vPosition");
    GLuint textureId  = glGetUniformLocation(mFboProId, "RenderTex");

    glEnableVertexAttribArray(vertexId);
    glBindBuffer(GL_ARRAY_BUFFER, mVertexBuf);
    glVertexAttribPointer(vertexId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);

    glEnableVertexAttribArray(texCoordId);
    glBindBuffer(GL_ARRAY_BUFFER, mTexCoordBuf);
    glVertexAttribPointer(texCoordId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);

    glBindTexture(GL_TEXTURE_2D, mTexId);
    glUniform1i(textureId, 0);
    glDrawArrays(GL_QUADS, 0, 4);
    glDisableVertexAttribArray(texCoordId);
    glDisableVertexAttribArray(vertexId);
    return LIBGLTF_SUCCESS;
}

int RenderWithFBO::createAndBindFbo(int width, int height)
{
    if (0 != mFboId)
    {
        return LIBGLTF_SUCCESS;
    }

    mFboProId = loadFboShader(FBOVERT, FBOFRAG);
    if(mFboProId == 0)
    {
        return LIBGLTF_SHADER_ERROR;
    }
    createRenderObj(width, height, mRboId);
    createTextureObj(width, height, mTexId);
    int status = createFrameBufferObj(mFboId, mTexId, mRboId);
    if (LIBGLTF_SUCCESS != status)
        return status;

    GLfloat coordVertices[] =
    {
        0.0f, 0.0f,
        1.0f, 0.0f,
        1.0f, 1.0f,
        0.0f, 1.0f,
    };
    GLfloat squareVertices[] =
    {
        -1.0f, -1.0f,
         1.0f, -1.0f,
         1.0f,  1.0f,
        -1.0f,  1.0f,
    };
    setBufferForFbo(mTexCoordBuf, mVertexBuf, coordVertices,
                    sizeof(coordVertices), squareVertices,
                    sizeof(squareVertices));
    return LIBGLTF_SUCCESS;
}

void RenderWithFBO::releaseFbo()
{
    if (0 != mFboId)
    {
        glDeleteFramebuffers(1, &mFboId);
        glDeleteRenderbuffers(1, &mRboId);
        glDeleteTextures(1, &mTexId);
        mFboId = 0;
    }
    if (0 != mShotFboId)
    {
        glDeleteFramebuffers(1, &mShotFboId);
        glDeleteRenderbuffers(1, &mShotRboId);
        glDeleteTextures(1, &mShotTexId);
        mShotFboId = 0;
    }
}

RenderPrimitive::RenderPrimitive()
    : mIndicesDataType(DataType_UNKNOW)
    , mGlobalmatrix(1.0f)
    , mLocalMatrix(1.0f)
    , mParentMatrixes()
    , pMaterial(0)
    , pNode(0)
    , mVerterCount(0)
    , mIndicesCount(0)
    , mVertexBuffer(0)
    , mNormalBuffer(0)
    , mTexCoordBuffer(0)
    , mJointBuffer(0)
    , mWeightBuffer(0)
    , mIndicesBuffer(0)
    , mVertexBufferData(0)
    , mIndiceBufferData(0)
    , mSortedIndiceBufferData(0)
    , mIndiceBufferLen(0)
    , mEyeSpaceVerts()
    , mPrimitiveZ()
    , mIsPolygonSorted(false)
{
}

RenderPrimitive::~RenderPrimitive()
{
    if (0 != mVertexBufferData)
    {
        delete[] mVertexBufferData;
    }
    if (0 != mIndiceBufferData)
    {
        delete[] mIndiceBufferData;
    }
    if (0 != mSortedIndiceBufferData)
    {
        delete[] mSortedIndiceBufferData;
    }
    glDeleteBuffers(1, &mVertexBuffer);
    glDeleteBuffers(1, &mNormalBuffer);
    glDeleteBuffers(1, &mTexCoordBuffer);
    glDeleteBuffers(1, &mIndicesBuffer);
}

Node* RenderPrimitive::getNode()
{
    return this->pNode;
}

void RenderPrimitive::setNode(Node* node)
{
    this->pNode = node;
}

Material* RenderPrimitive::getMaterial()
{
    return this->pMaterial;
}

void RenderPrimitive::setMaterial(Material* pMaterial)
{
    this->pMaterial = pMaterial;
}

unsigned int RenderPrimitive::getVerterCount()
{
    return this->mVerterCount;
}

void RenderPrimitive::setVerterCount(unsigned int count)
{
    this->mVerterCount = count;
}

unsigned int RenderPrimitive::getIndicesCount()
{
    return this->mIndicesCount;
}

void RenderPrimitive::setIndicesCount(unsigned int count)
{
    this->mIndicesCount = count;
}

DataType_E RenderPrimitive::getIndicesDataType()
{
    return this->mIndicesDataType;
}

void RenderPrimitive::setIndicesDataType(DataType_E type)
{
    this->mIndicesDataType = type;
}

void RenderPrimitive::copyIndiceBufferData(const char* srcBuf,
                                           unsigned int bufSize)
{
    if (0 == mIndiceBufferData)
    {
        mIndiceBufferData = new char[bufSize];
        memcpy(mIndiceBufferData,srcBuf,bufSize);
        mIndiceBufferLen = bufSize;
    }
    if (0 == mSortedIndiceBufferData)
    {
        mSortedIndiceBufferData = new char[bufSize];
    }
}
void RenderPrimitive::polyonSorting(glm::mat4& viewMatrix)
{
    mEyeSpaceVerts.resize(mVerterCount);
    glm::vec3* pVertices = (glm::vec3*)mVertexBufferData;
    for (size_t i = 0; i < mVerterCount; ++i)
    {
        mEyeSpaceVerts[i] = pVertices[i].x * viewMatrix[2][0]
                          + pVertices[i].y * viewMatrix[2][1]
                          + pVertices[i].z * viewMatrix[2][2]
                          + viewMatrix[2][3];
    }
    sortIndices();
}
void RenderPrimitive::sortIndices()
{
    mPrimitiveZ.resize(mIndicesCount / 3);
    if (mPrimitiveZ.empty())
    {
        mIsPolygonSorted = false;
        return;
    }
    if (DataType_UNSIGNED_SHORT == mIndicesDataType)
    {
        unsigned short* indexBuf = (unsigned short*)mIndiceBufferData;
        for (unsigned iz = 0; iz < mIndicesCount / 3; ++iz)
        {
            mPrimitiveZ[iz].mZ = (float)(mEyeSpaceVerts[indexBuf[iz * 3]]
                               + mEyeSpaceVerts[indexBuf[iz * 3 + 1]]
                               + mEyeSpaceVerts[indexBuf[iz * 3 + 2]]);
            mPrimitiveZ[iz].mPrimitiveIndex = iz;
        }
        std::sort(mPrimitiveZ.begin(), mPrimitiveZ.end(),
                  SorterBackToFront());
        unsigned short* sortedIndexBuf =
            (unsigned short*)mSortedIndiceBufferData;
        for (unsigned int i = 0; i < mPrimitiveZ.size(); ++i)
        {
            sortedIndexBuf[i * 3] =
                indexBuf[mPrimitiveZ[i].mPrimitiveIndex * 3];
            sortedIndexBuf[i * 3 + 1] =
                indexBuf[mPrimitiveZ[i].mPrimitiveIndex * 3 + 1];
            sortedIndexBuf[i * 3 + 2] =
                indexBuf[mPrimitiveZ[i].mPrimitiveIndex * 3 + 2];
        }
        mIsPolygonSorted = true;
    }
    else if (DataType_UNSIGNED_INT == mIndicesDataType)
    {
        unsigned int* indexBuf = (unsigned int*)mIndiceBufferData;
        for (unsigned iz = 0; iz < mIndicesCount / 3; ++iz)
        {
            mPrimitiveZ[iz].mZ = (float)(mEyeSpaceVerts[indexBuf[iz * 3]]
                               + mEyeSpaceVerts[indexBuf[iz * 3 + 1]]
                               + mEyeSpaceVerts[indexBuf[iz * 3 + 2]]);
            mPrimitiveZ[iz].mPrimitiveIndex = iz;
        }
        std::sort(mPrimitiveZ.begin(), mPrimitiveZ.end(),
                  SorterBackToFront());
        unsigned int* sortedIndexBuf = (unsigned int *)mSortedIndiceBufferData;
        for (unsigned int i= 0; i < mPrimitiveZ.size(); ++i)
        {
            sortedIndexBuf[i * 3] =
                indexBuf[ mPrimitiveZ[i].mPrimitiveIndex * 3];
            sortedIndexBuf[i * 3 + 1] =
                indexBuf[ mPrimitiveZ[i].mPrimitiveIndex * 3 + 1];
            sortedIndexBuf[i * 3 + 2] =
                indexBuf[ mPrimitiveZ[i].mPrimitiveIndex * 3 + 2];
        }
        mIsPolygonSorted = true;
    }
    else
    {
        mIsPolygonSorted = false;
    }
}

void RenderPrimitive::bindSortedIndicesBuf()
{
    if (mIsPolygonSorted)
    {
        glDeleteBuffers(1, &mIndicesBuffer);
        glGenBuffers(1, &mIndicesBuffer);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mIndicesBuffer);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, mIndiceBufferLen,
                     mSortedIndiceBufferData, GL_STATIC_DRAW);
    }
}

void RenderPrimitive::copyVertexBufferData(const char* srcBuf,
                                           unsigned int bufSize)
{
    if (0 == mVertexBufferData)
    {
        mVertexBufferData = new char[bufSize];
        memcpy(mVertexBufferData, srcBuf, bufSize);
    }
}

unsigned int RenderPrimitive::getVertexBuffer()
{
    return this->mVertexBuffer;
}

void RenderPrimitive::setVertexBuffer(unsigned int bufferId)
{
    this->mVertexBuffer = bufferId;
}

unsigned int RenderPrimitive::getNormalBuffer()
{
    return this->mNormalBuffer;
}

void RenderPrimitive::setNormalBuffer(unsigned int bufferId)
{
    this->mNormalBuffer = bufferId;
}

unsigned int RenderPrimitive::getTexCoordBuffer()
{
    return this->mTexCoordBuffer;
}

void RenderPrimitive::setTexCoordBuffer(unsigned int bufferId)
{
    this->mTexCoordBuffer = bufferId;
}

unsigned int RenderPrimitive::getJointBuffer()
{
    return this->mJointBuffer;
}

void RenderPrimitive::setJointBuffer(unsigned int bufferId)
{
    this->mJointBuffer = bufferId;
}

unsigned int RenderPrimitive::getWeightBuffer()
{
    return this->mWeightBuffer;
}

void RenderPrimitive::setWeightBuffer(unsigned int bufferId)
{
    this->mWeightBuffer = bufferId;
}

unsigned int RenderPrimitive::getIndicesBuffer()
{
    return this->mIndicesBuffer;
}

void RenderPrimitive::setIndicesBuffer(unsigned int bufferId)
{
    this->mIndicesBuffer = bufferId;
}



RenderShader::RenderShader()
    : mPrimitiveVec()
    , pTechnique(0)
{
}

RenderShader::~RenderShader()
{
    for (unsigned int i = 0, size = mPrimitiveVec.size();
         i < size; ++i)
    {
        delete mPrimitiveVec[i];
    }
    mPrimitiveVec.clear();
}

Technique* RenderShader::getTechnique()
{
    return this->pTechnique;
}

void RenderShader::setTechnique(Technique* pTechnique)
{
    this->pTechnique = pTechnique;
}

void RenderShader::pushRenderPrim(RenderPrimitive* p)
{
    (this->mPrimitiveVec).push_back(p);
}

RenderPrimitive* RenderShader::getRenderPrim(unsigned int index)
{
    if (this->mPrimitiveVec.size() <= index)
    {
        return 0;
    }
    return this->mPrimitiveVec[index];
}

unsigned int RenderShader::getRenderPrimSize()
{
    return (this->mPrimitiveVec).size();
}

ShaderProgram RenderScene::mShaderProgram = ShaderProgram();

RenderScene::RenderScene()
    : pCamera(0)
    , cCamera(0)
    , vCamera()
    , vCameraIndex()
    , roCameraX(1.0f)
    , roCameraY(1.0f)
    , tagView(1.0f)
    , IsAerialMode(false)
    , bAerialView(false)
    , oldViewMatrix(1.0f)
    , oldTime(0.0)
    , flyInfo(1.0f)
    , flyTime(0.0)
    , bFlyCamera(false)
    , bAnimation(true)
    , initCameraLocalMatrix(1.0f)
    , initCameraGlobalMatrix(1.0f)
    , pLight(0)
    , light0Transfrom(1.0f)
    , pTempMatrix(0)
    , mAnimationPlay(false)
    , mAnimationLoop(true)
    , mCurrentTime(0)
    , mLastPlaying(0)
    , mUpdateTimeOut(0)
    , mDuration(0)
    , mCameraMoveTime(0)
    , mPreCameraTime(0)
    , mShaderVec()
    , pScene(0)
    , mLoadJson()
    , mBindBufferMap()
    , mCurrentViewport()
    , fbo()
    , mEnableTransparency(false)
    , mEnableRotation(false)
    , mRotationLastTime(0.0)
    , mLastModelView(glm::mat4(0.0))
    , bIsTimeAvailable(false)
#if ENABLE_FPS
    , pFPSCounter(0)
#endif
#if ENABLE_TIMER
    , pTimer(new Timer())
#endif
#if LOAD_ONCE
    , mCurrentImage()
    , mCurrentTextNumber(0)
#endif
{
}

RenderScene::~RenderScene()
{
#if ENABLE_FPS
    if (0 != pFPSCounter)
    {
        delete pFPSCounter;
    }
#endif
#if ENABLE_TIMER
    if (0 != pTimer)
    {
        delete pTimer;
    }
#endif
    if (0 != pLight)
    {
        delete pLight;
    }
    if (0 != pCamera)
    {
        delete pCamera;
    }
    for (unsigned int i = 0, size = mShaderVec.size();
         i < size; ++i)
    {

         delete mShaderVec[i];
    }
    mShaderVec.clear();
    mBindBufferMap.clear();
    if (0 != pScene)
    {
        delete pScene;
    }
    delete[] pTempMatrix;
}

int RenderScene::loadScene(const std::string& directory,
                           const std::string& file)
{
    mLoadJson.setJsonInfo(directory, file);
    if (!mLoadJson.parseJsonFile())
    {
        return LIBGLTF_PARSE_JSON_ERROR;
    }
    int iStatus = mLoadJson.parseScene();
    return iStatus;
}

int RenderScene::loadScene()
{
    try
    {
        int iStatus = mLoadJson.parseScene();
        return iStatus;
    }
    catch (boost::property_tree::ptree_error& e)
    {
        fprintf(stderr, "%s\n", e.what());
        return LIBGLTF_PARSE_JSON_ERROR;
    }
    catch (boost::exception const&)
    {
        fprintf(stderr, "%s\n", "unknown boost error");
        return LIBGLTF_UNKNOWN_BOOST_ERROR;
    }
}

glTFHandle* RenderScene::initScene(const glTFFile* jsonfile)
{
    if (0 == jsonfile)
    {
        return 0;
    }
    string tmpStr(jsonfile->buffer, jsonfile->size);
    mLoadJson.setJsonInfo(tmpStr);
    if (!mLoadJson.parseJsonFile())
    {
        return 0;
    }
    string tmpFileName(jsonfile->filename);
    glTFHandle* handle = mLoadJson.getFileNameInJson(tmpFileName);
    pScene = new Scene();
    pScene->setGltfHandle(handle);
    mLoadJson.setScene(pScene);
    return handle;
}

unsigned int RenderScene::bindAttribute(const Attribute* pAttr)
{
    unsigned int bufferId;
    glGenBuffers(1, &bufferId);
    glBindBuffer(GL_ARRAY_BUFFER, bufferId);
    glBufferData(GL_ARRAY_BUFFER,
                 pAttr->getDataCount() * pAttr->getByteStride(),
                 pAttr->getAttributeData(), GL_STATIC_DRAW);
    return bufferId;
}

unsigned int RenderScene::bindIndices(const Attribute* pAttr)
{
    unsigned int bufferId;
    glGenBuffers(1, &bufferId);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, bufferId);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER,
                 pAttr->getDataCount() * pAttr->getByteStride(),
                 pAttr->getAttributeData(), GL_STATIC_DRAW);
    return bufferId;
}

void RenderScene::constructShader()
{
    for (unsigned int i = 0, size = pScene->getTechSize();
         i < size; ++i)
    {
        RenderShader* pRenderShader = new RenderShader();
        pRenderShader->setTechnique(pScene->getTechnique(i));
        mShaderVec.push_back(pRenderShader);
    }
}

void RenderScene::constructMesh(const std::string& meshName, Node* pNode)
{
    const Mesh* pMesh = pScene->findMesh(meshName);
    for (unsigned int i = 0, size = pMesh->getPrimitiveVecSize();
         i < size; ++i)
    {
        constructPrimitive(pMesh->getPrimitiveVec(i), pNode);
    }
}

void RenderScene::constructPrimitive(const Primitives* pPrimitive,
                                     Node* pNode)
{
    const std::string& materialIndex = pPrimitive->getMaterialIndex();
    Material* pMaterial = pScene->findMaterial(materialIndex);
    if (0 != pMaterial)
    {
        for (unsigned int i = 0, size = mShaderVec.size();
             i < size; ++i)
        {
            RenderShader* pRenderShader = mShaderVec[i];
            Technique* pTechnique = pRenderShader->getTechnique();
            if (pMaterial->getTechniqueId() == pTechnique->getTechId())
            {
                RenderPrimitive* pRenderPrimitive = new RenderPrimitive();
                bindAttributeBuffer(pPrimitive, pRenderPrimitive);
                pRenderPrimitive->setMaterial(pMaterial);
                pRenderPrimitive->setNode(pNode);

                pRenderShader->pushRenderPrim(pRenderPrimitive);
                break;
            }
        }
    }
}

void RenderScene::bindAttributeBuffer(const Primitives* pPrimitive,
                                      RenderPrimitive* pRP)
{
    const Attribute* pAttr;
    BindBufferInfo bindBufferInfo;
    std::map<std::string, BindBufferInfo>::const_iterator it;

    const std::string& posId = pPrimitive->getAttributeIndex("POSITION");
    it = mBindBufferMap.find(posId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(posId);
        if (0 != pAttr)
        {
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            bindBufferInfo.mDataCount = pAttr->getDataCount();
            bindBufferInfo.mBufferLen = pAttr->getDataCount()
                                        * pAttr->getByteStride();
            bindBufferInfo.mSrcData = pAttr->getAttributeData();
            pRP->setVertexBuffer(bindBufferInfo.mBufferId);
            pRP->setVerterCount(bindBufferInfo.mDataCount);
            pRP->copyVertexBufferData(bindBufferInfo.mSrcData,
                                      bindBufferInfo.mBufferLen);
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (posId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setVertexBuffer(it->second.mBufferId);
        pRP->setVerterCount(it->second.mDataCount);
        pRP->copyVertexBufferData(it->second.mSrcData,
                                  it->second.mBufferLen);
    }

    const std::string& NorId = pPrimitive->getAttributeIndex("NORMAL");
    it = mBindBufferMap.find(NorId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(NorId);
        if (0 != pAttr)
        {
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            pRP->setNormalBuffer(bindBufferInfo.mBufferId);
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (NorId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setNormalBuffer(it->second.mBufferId);
    }

    const std::string& TexId = pPrimitive->getAttributeIndex("TEXCOORD_0");
    it = mBindBufferMap.find(TexId);
    if (mBindBufferMap.end() == it)
    {
        pAttr = pScene->findAttribute(TexId);
        if (0 != pAttr)
        {
            float* pData = (float*)pAttr->getAttributeData();
            for (unsigned int i = 0, size = pAttr->getDataCount();
                 i < size; i++)
            {
                pData[i * 2 + 1] = 1 - pData[i * 2 + 1];
            }
            bindBufferInfo.mBufferId = bindAttribute(pAttr);
            pRP->setTexCoordBuffer(bindBufferInfo.mBufferId);
            mBindBufferMap.insert(
                std::map<std::string, BindBufferInfo>::value_type
                (TexId, bindBufferInfo));
        }
    }
    else
    {
        pRP->setTexCoordBuffer(it->second.mBufferId);
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("JOINT"));
    if (0 != pAttr)
    {
        pRP->setJointBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getAttributeIndex("WEIGHT"));
    if (0 != pAttr)
    {
        pRP->setWeightBuffer(bindAttribute(pAttr));
    }

    pAttr = pScene->findAttribute(pPrimitive->getIndicesIndex());
    if (0 != pAttr)
    {
        pRP->setIndicesBuffer(bindIndices(pAttr));
        pRP->setIndicesCount(pAttr->getDataCount());
        pRP->setIndicesDataType(pAttr->getDataType());
        pRP->copyIndiceBufferData(pAttr->getAttributeData(),
                                  pAttr->getDataCount() *
                                  pAttr->getByteStride());
    }
}

void RenderScene::getCameraIndex(Node* pNode)
{
    Node* pChild = 0;
    for (unsigned int i = 0,size = pNode->getChildNodeSize();
         i < size; ++i)
    {
        pChild = pNode->getChildNode(i);
        if (!pChild->getCameraIndex().empty())
        {
            vCameraIndex.push_back(pChild->getCameraIndex());
        }
        getCameraIndex(pChild);
    }
    return;
}
void RenderScene::initOpengl()
{
    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glDepthFunc(GL_LESS);
}

void RenderScene::createDefaultCamera()
{
    glm::vec3 vDelta = pScene->getVertexMax() - pScene->getVertexMin();
    glm::vec3 vHalf =  pScene->getVertexMin() + (vDelta / 2.0f);

    float fPos = sqrt(vDelta.x * vDelta.x
                    + vDelta.y * vDelta.y
                    + vDelta.z * vDelta.z);
    if (fPos >= 3000)
    {
        pCamera = new CPhysicalCamera(
            glm::vec3(vHalf.x, 0.0, fPos),
            vHalf,
            glm::vec3(0.0f, 1.0f, 0.0f));
    }
    else
    {
        pCamera = new CPhysicalCamera(glm::vec3(0.0, 0.0, 1.5 * fPos),
                                      vHalf,
                                      glm::vec3(0.0f, 1.0f, 0.0f));
    }
    pCamera->mLook = glm::lookAt(pCamera->vEye, pCamera->vView, pCamera->vUp);
    tagView = pCamera->vView;
    pCamera->flength = fPos;
    pCamera->vModelCenterPos = vHalf;

    getCameraIndex(pScene->getRootNode());
    if (vCameraIndex.size() > 0)
    {
        std::vector<std::string>::iterator it;
        it = vCameraIndex.begin();
        cCamera = pScene->findCamera(*it);
    }
    if (pScene->getUseCameraInJson())
    {
        pCamera->setPerspective(glm::perspective(cCamera->getXFov(),
                                                 cCamera->getAspectRatio(),
                                                 cCamera->getNear(),
                                                 cCamera->getFar()));
    }
    else
    {
        pCamera->setPerspective(glm::perspective(37.8492f, 1.5f, 1.0f,
                                                 500000.0f));
    }
}

void RenderScene::initNodeTree(Node* pNode, const glm::mat4& matrix,
                               bool parentJointFlag, bool updateFlag)
{
    glm::mat4 globalMatrix;
    bool currJointFlag = pNode->getJointFlag();
    if (!parentJointFlag && currJointFlag)
    {
        globalMatrix = pNode->getLocalMatrix();
    }
    else
    {
        globalMatrix = matrix * pNode->getLocalMatrix();
    }

    if (!pNode->getMatrixFlag())
    {
        updateFlag = true;
    }
    pNode->setUpdateFlag(updateFlag);

    //if (pNode->getSkinIndex().empty())
    {
        pNode->setGlobalMatrix(globalMatrix);
    }

    for (unsigned int i = 0, nSize = pNode->getChildNodeSize();
         i < nSize; ++i)
    {
        initNodeTree(pNode->getChildNode(i), globalMatrix,
                     currJointFlag, updateFlag);
    }
}

int RenderScene::initRender()
{
    initOpengl();

    int iResult = loadScene();
    if (iResult != LIBGLTF_SUCCESS)
    {
        return iResult;
    }
    pTempMatrix = new glm::mat4[MAX_BONE_SIZE];

    Node* pRootNode = pScene->getRootNode();

    createDefaultCamera();
    trackball(pCamera->curquatY, 0.0f, 0.0f, 0.0f, 0.0f);
    trackball(pCamera->curquatX, 0.0f, 0.0f, 0.0f, 0.0f);
    trackball(pCamera->RotateCurquatY, 0.0f, 0.0f, 0.0f, 0.0f);
    trackball(pCamera->RotateCurquatX, 0.0f, 0.0f, 0.0f, 0.0f);
    constructShader();
    initNodeTree(pRootNode, pRootNode->getGlobalMatrix(), false, false);
    if (pScene->getUseCameraInJson())
    {
        initCameraLocalMatrix = cCamera->getCameraNode()->getLocalMatrix();
        initCameraGlobalMatrix = cCamera->getCameraNode()->getGlobalMatrix();
    }

    for (unsigned int i = 0, nSize = pScene->getNodeSize();
         i < nSize; ++i)
    {
        Node* pNode = pScene->getNode(i);

        if (0 != pScene->getAnimationCount())
        {
            const std::string& nodeName = pNode->getNodeName();
            pNode->setAnimPoint(pScene->findAnimation(nodeName));
        }

        const std::string& skinIndex = pNode->getSkinIndex();
        if (!skinIndex.empty())
        {
            Node* pPareBone = 0;
            Node* pBone = 0;
            pPareBone = findNodeByName(pRootNode, pNode->getSkeleIndex());
            for (unsigned int i = 0, size = pScene->getSkinSize();
                 i < size; ++i)
            {
                Skin* pSkin = pScene->getSkin(i);
                if (pSkin->getSkinName() == skinIndex)
                {
                    pNode->setSkinPoint(pSkin);
                    for (unsigned int i = 0, jointCount = pSkin->getBoneIdSize();
                         i < jointCount; i++)
                    {
                        pBone = findNodeByJoint(pPareBone, pSkin->getBoneId(i));
                        pNode->pushBoneNode(pBone);
                    }
                    break;
                }
                continue;
            }
        }

        if (pNode->hasMesh())
        {
            for (unsigned int j = 0, size = pNode->getMeshIndexSize();
                 j < size; ++j)
            {
                constructMesh(pNode->getMeshIndex(j), pNode);
            }
        }

    }
    pScene->clearAttributeMap();

    this->mDuration = pScene->getDuration();

    return LIBGLTF_SUCCESS;
}
#if ENABLE_FPS
void RenderScene::initFPS()
{
    if (0 != pFPSCounter)
    {
        delete pFPSCounter;
    }
    pFPSCounter = new FPSCounter(&mShaderProgram);
}
#else
void RenderScene::initFPS(unsigned char* /*fontBuf*/,unsigned int /*bufSize*/)
{
    return;
}
#endif
void RenderScene::renderPrimitive(RenderPrimitive* pPrimitive,
                                  unsigned int progId)
{
    upLoadMatrixInfo(progId, pPrimitive);
    if (mEnableTransparency)
    {
        pPrimitive->bindSortedIndicesBuf();
    }

    {
        TRACE_TIME;
        upLoadUniform(progId, pPrimitive);
    }

    {
        TRACE_TIME;
        upLoadAttribute(progId, pPrimitive);
    }

    {
        TRACE_TIME;
        upLoadAnimation(progId, pPrimitive);
    }

    {
        TRACE_TIME;
        drawTriangle(pPrimitive);
    }
}

void RenderScene::upLoadMatrixInfo(unsigned int progId,
                                   RenderPrimitive* pPrimitive)
{
    Node* pNode = pPrimitive->getNode();
    glm::mat4& tempMatrix = pNode->getGlobalMatrix();
    glm::mat4& lookMatrix = pCamera->getLook();

    {
        TRACE_TIME;
        mShaderProgram.setUniform(progId, "u_modelViewMatrix",
            pCamera->getModelViewMatrix(lookMatrix,
                                        tempMatrix,
                                        mCameraMoveTime,
                                        mPreCameraTime));
        mShaderProgram.setUniform(progId, "u_normalMatrix",
            glm::mat3(lookMatrix) *
            glm::transpose(glm::inverse(glm::mat3(tempMatrix))));
        mShaderProgram.setUniform(progId, "u_projectionMatrix",
                                  pCamera->getPerspective());
        mShaderProgram.setUniform(progId, "M", tempMatrix);
        mShaderProgram.setUniform(progId, "V", lookMatrix);
    }
}

bool RenderScene::upLoadTechProperty()
{
    glEnable(GL_BLEND);
    glBlendEquation(GL_FUNC_ADD);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glDepthMask(GL_TRUE);
    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
    return true;
}

void RenderScene::upLoadTechPropertyOfJsonFile(Technique* pTech)
{
    if (1 == pTech->getTechState()->blendEnable )
    {
        glEnable(GL_BLEND);
    }
    else
    {
        glDisable(GL_BLEND);
    }

    if (0 != pTech->getTechState()->blendEquation)
    {
        glBlendEquation(pTech->getTechState()->blendEquation);
        glBlendFunc(pTech->getTechState()->blendFuncSfactor,
                    pTech->getTechState()->blendFuncDfactor);
    }

    if (1 == pTech->getTechState()->cullFaceEnable)
    {
        glEnable(GL_CULL_FACE);
    }
    else
    {
        glDisable(GL_CULL_FACE);
    }

    if (1 == pTech->getTechState()->depthMask)
    {
        glDepthMask(GL_TRUE);
    }
    else
    {
        glDepthMask(GL_FALSE);
    }

    if (1 == pTech->getTechState()->depthTestEnable)
    {
        glEnable(GL_DEPTH_TEST);
    }
    else
    {
        glDisable(GL_DEPTH_TEST);
    }
}

void RenderScene::upLoadTechInfo(unsigned int progId, Technique* pTech)
{
    if (mEnableTransparency)
    {
        upLoadTechProperty();
    }
    else
    {
        upLoadTechPropertyOfJsonFile(pTech);
    }

    std::vector<techLight*> vecLight = pTech->poptLight();
    std::vector<techLight*>::const_iterator it = vecLight.begin();
    std::vector<techLight*>::const_iterator itEnd = vecLight.end();
    for (; it != itEnd; ++it)
    {
        if ((*it)->mSource.empty())
        {
            switch ((*it)->type)
            {
            case DataType_FLOAT:
                mShaderProgram.setUniform(progId, (*it)->mName.c_str(),
                    (*it)->floatValue);
                break;
            case DataType_FLOAT_VEC3:
                mShaderProgram.setUniform(progId, (*it)->mName.c_str(),
                    (*it)->vecValue);
                break;
            default: break;
            }
        }
        else
        {
            Node* pNode = pScene->findLightNodeMap((*it)->mSource);
            mShaderProgram.setUniform(progId, (*it)->mName.c_str(),
                                      pCamera->getLook() *
                                      pNode->getGlobalMatrix());
        }
    }
}

void RenderScene::upLoadAnimation(unsigned int progId,
                                  RenderPrimitive* pPrimitive)
{
    Node* pNode = pPrimitive->getNode();
    Skin* pSkin = pNode->getSkinPoint();
    if (0 == pSkin)
    {
        return;
    }

    unsigned int count = pSkin->getBindMatrixCount();
    memcpy(pTempMatrix, pSkin->getBindMatrix(), sizeof(glm::mat4)*count);

    Node* pBone;
    for (unsigned int i = 0; i < count; ++i)
    {
        pBone = pNode->getBoneNode(i);
        if (0 != pBone)
            pTempMatrix[i] = pBone->getGlobalMatrix() * pTempMatrix[i];
    }

    glUniformMatrix4fv(glGetUniformLocation(progId, "u_jointMat"),
                       count, false, (GLfloat*)pTempMatrix);
}

void RenderScene::updateAnimInfo(Node* pNode)
{
    Animation* pAnimation = pNode->getAnimPoint();
    if (0 == pAnimation)
    {
        return;
    }

    double time = fmod(mCurrentTime, pAnimation->getDuration());
    glm::mat4 localMatrix = pAnimation->findTimeValue(time);
    if (pAnimation->getChannelBits() == ROTATE_CHANNEL)
        localMatrix = pNode->getTranslate() * localMatrix * pNode->getScale();
    pNode->setLocalMatrix(localMatrix);
}

void RenderScene::updateNodeMatrix(Node* pNode, const glm::mat4& matrix,
                                   bool jointFlag)
{
    glm::mat4 globalMatrix;
    bool flag = pNode->getJointFlag();

    if (pNode->getUpdateFlag())
    {
        updateAnimInfo(pNode);
        if (!jointFlag && flag)
        {
            globalMatrix = pNode->getLocalMatrix();
        }
        else
        {
            globalMatrix = matrix * pNode->getLocalMatrix();
        }

        if (pNode->getSkinIndex().empty())
        {
            pNode->setGlobalMatrix(globalMatrix);
        }
    }
    else
    {
        globalMatrix = pNode->getGlobalMatrix();
    }

    for (unsigned int i = 0, nSize = pNode->getChildNodeSize();
         i < nSize; ++i)
    {
        updateNodeMatrix(pNode->getChildNode(i), globalMatrix, flag);
    }
}

void RenderScene::updateViewMatrix()
{
    glm::mat4 viewMatrixTemp;
    if (pScene->getUseCameraInJson())
    {
        Node* pNode = cCamera->getCameraNode();
        glm::vec3 vMoveTemp;
        vMoveTemp = pCamera->vMove;
        if (pCamera->bMoveCameraModel)
        {
            viewMatrixTemp = glm::translate(pNode->getLocalMatrix(),
                                            vMoveTemp);
            pNode->setLocalMatrix(viewMatrixTemp);
            viewMatrixTemp = pNode->getParentNode()->getGlobalMatrix() *
                             viewMatrixTemp;
            pNode->setGlobalMatrix(viewMatrixTemp);
        }
        else
        {
            viewMatrixTemp = pNode->getGlobalMatrix();
        }
        viewMatrixTemp = glm::inverse(viewMatrixTemp);
    }
    else
    {
        if (pCamera->bMoveCameraModel)
        {
            pCamera->mLook = glm::translate(pCamera->mLook,
                                            glm::vec3(-pCamera->vMove.x,
                                                       0.0,
                                                      -pCamera->vMove.z));
        }
        viewMatrixTemp = pCamera->mLook;
    }
    //Update Camera Matrix
    if (pCamera->bRotateCameraModel)
    {
        if( bAerialView )
        {
            pCamera->buildRotMatrix(pCamera->mTrackBallY, pCamera->curquatY);
            pCamera->buildRotMatrix(pCamera->mTrackBallX, pCamera->curquatX);
        }
        else
        {
            pCamera->buildRotMatrix(roCameraY, pCamera->RotateCurquatY);
            pCamera->buildRotMatrix(roCameraX, pCamera->RotateCurquatX);
        }
    }
    viewMatrixTemp *= pCamera->mTrackBallY;
    viewMatrixTemp *= pCamera->mTrackBallX;
    viewMatrixTemp = roCameraY * roCameraX * viewMatrixTemp;
    pCamera->setLook(viewMatrixTemp);
}

void RenderScene::updateFlyCamera()
{
    static bool startFly = false;
    pCamera->bRotateCameraModel = false;
    if (flyTime<=0)
    {
        bFlyCamera = false;
        startFly = false;
        return;
    }
    glm::mat4 viewMatrix = pCamera->getLook();
    if (!startFly)
    {
        startFly = true;
        oldTime = libgltf::time::getCurrentTime();
        pCamera->mLook = viewMatrix;
        pCamera->mTrackBallY = glm::mat4(1.0f);
        pCamera->mTrackBallX = glm::mat4(1.0f);
        roCameraY = glm::mat4(1.0f);
        roCameraX = glm::mat4(1.0f);
        trackball(pCamera->RotateCurquatY, 0.0f, 0.0f, 0.0f, 0.0f);
        trackball(pCamera->RotateCurquatX, 0.0f, 0.0f, 0.0f, 0.0f);
        trackball(pCamera->curquatY, 0.0f, 0.0f, 0.0f, 0.0f);
        trackball(pCamera->curquatX, 0.0f, 0.0f, 0.0f, 0.0f);
        return;
    }
    double newTime = libgltf::time::getCurrentTime();
    double timeDifference = libgltf::time::diffTime(newTime,oldTime)/1000.0;
    flyTime -= timeDifference;
    viewMatrix += flyInfo * (float)timeDifference;
    pCamera->mLook = viewMatrix;
}
void RenderScene::upLoadUniform(unsigned int progId,
                                RenderPrimitive* pPrimitive)
{
    int textureCount = 0;
    Material* pMaterial = pPrimitive->getMaterial();
    for (unsigned int i = 0, size = pMaterial->getMaterialProperSize();
         i < size; ++i)
    {
        const MaterialProperty* pProper = pMaterial->getMaterialProper(i);
        switch (pProper->getDataType())
        {
        case DataType_FLOAT:
            mShaderProgram.setUniform(progId, pProper->getPropertyName().c_str(),
                                      (float*)(pProper->getPropertyData()));
            break;
        case DataType_FLOAT_VEC2:
            mShaderProgram.setUniform(progId, pProper->getPropertyName().c_str(),
                                      (glm::vec2*)(pProper->getPropertyData()));
            break;
        case DataType_FLOAT_VEC3:
            mShaderProgram.setUniform(progId, pProper->getPropertyName().c_str(),
                                      (glm::vec3*)(pProper->getPropertyData()));
            break;
        case DataType_FLOAT_VEC4:
            mShaderProgram.setUniform(progId, pProper->getPropertyName().c_str(),
                                      (glm::vec4*)(pProper->getPropertyData()));
            break;
        case DataType_FLOAT_MAT3:
            mShaderProgram.setUniform(progId, pProper->getPropertyName().c_str(),
                                      (glm::mat3*)(pProper->getPropertyData()));
            break;
        case DataType_FLOAT_MAT4:
            mShaderProgram.setUniform(progId, pProper->getPropertyName().c_str(),
                                      (glm::mat4*)(pProper->getPropertyData()));
            break;
        case DataType_SAMPLER_2D:
#if LOAD_ONCE
            if (mCurrentImage == pProper->getImagePath() &&
                textureCount == mCurrentTextNumber)
            {
                break;
            }
            mCurrentImage = pProper->getImagePath();
            mCurrentTextNumber = textureCount;
#endif
            mShaderProgram.setUniform(progId, pProper->getPropertyName().c_str(),
                                      textureCount);
            pScene->findTexture(pProper->getImagePath())->bindTexture(textureCount);
            textureCount++;
            break;
        default:
            break;
        }
    }
    return;
}

void RenderScene::upLoadAttribute(unsigned int progId,
                                  RenderPrimitive* pPrimitive)
{
    int tmpId;
    if ((tmpId = glGetAttribLocation(progId, "a_position")) != -1)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getVertexBuffer());
        glVertexAttribPointer(tmpId, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }
    if ((tmpId = glGetAttribLocation(progId, "a_normal")) != -1)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getNormalBuffer());
        glVertexAttribPointer(tmpId, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }
    if ((tmpId = glGetAttribLocation(progId, "a_texcoord0")) != -1)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getTexCoordBuffer());
        glVertexAttribPointer(tmpId, 2, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }
    if ((tmpId = glGetAttribLocation(progId, "a_weight")) != -1)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getWeightBuffer());
        glVertexAttribPointer(tmpId, 4, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }
    if ((tmpId = glGetAttribLocation(progId, "a_joint")) != -1)
    {
        glEnableVertexAttribArray(tmpId);
        glBindBuffer(GL_ARRAY_BUFFER, pPrimitive->getJointBuffer());
        glVertexAttribPointer(tmpId, 4, GL_FLOAT, GL_FALSE, 0, (void*)0);
    }
    return;
}

void RenderScene::drawTriangle(RenderPrimitive* pPrimitive)
{
    if (pPrimitive->getIndicesCount() > 0)
    {
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, pPrimitive->getIndicesBuffer());
#if ENABLE_TIMER
        glFinish();
#endif
        glDrawElements(GL_TRIANGLES, pPrimitive->getIndicesCount(),
                       pPrimitive->getIndicesDataType(), 0);
#if ENABLE_TIMER
        glFinish();
#endif
    }
    else
    {
        glDrawArrays(GL_TRIANGLES, 0, pPrimitive->getVerterCount());
    }
}

int RenderScene::prepareRender(void* lpParam)
{
    int status;
#if ENABLE_FPS
    if (0 != pFPSCounter)
    {
        pFPSCounter->timeStamp();
    }
#endif
    {
        TRACE_TIME;
        status = initSSAAFrameBuf(((glTFViewport*)lpParam));
        if (LIBGLTF_SUCCESS != status)
            return status;
    }
    return 0;
}

int RenderScene::initSSAAFrameBuf(glTFViewport* pViewport)
{
    if (0 == pViewport->width)
        return LIBGLTF_INVALID_SIZE;

    // When viewport changes we need to release fbo and create a new one
    if( mCurrentViewport.x != pViewport->x || mCurrentViewport.y != pViewport->y ||
    mCurrentViewport.width != pViewport->width || mCurrentViewport.height != pViewport->height )
    {
        fbo.releaseFbo();
        mCurrentViewport = *pViewport;
    }

#if DEFAULT_VIEW
    unsigned int width  = SSAA * DEFAULT_VIEW;
    unsigned int height = (SSAA * DEFAULT_VIEW *
                           pViewport->height / pViewport->width);
#else
    unsigned int width  = SSAA * pViewport->width;
    unsigned int height = SSAA * pViewport->height;
#endif //DEFAULT_VIEW
    int status = fbo.createAndBindFbo(width, height);
    if (LIBGLTF_SUCCESS != status)
        return status;
    glBindFramebuffer(GL_FRAMEBUFFER, fbo.mFboId);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glViewport(pViewport->x, pViewport->y, width, height);
    return LIBGLTF_SUCCESS;
}

void RenderScene::render()
{
    setTimeForAnim();
    realRender();
}

void RenderScene::realRender()
{
    if (mCurrentTime >= mUpdateTimeOut || !bIsTimeAvailable)
    {
        if (this->mAnimationPlay)
        {
            Node* pRootNode = pScene->getRootNode();
            updateNodeMatrix(pRootNode, pRootNode->getGlobalMatrix(), false);
        }
        mUpdateTimeOut = mCurrentTime;
        updateViewMatrix();
        pCamera->bRotateCameraModel = false;
        pCamera->bMoveCameraModel = false;
    }
    else if(bAerialView && (pCamera->bRotateCameraModel || pCamera->bMoveCameraModel))
    {
        updateViewMatrix();
        pCamera->bRotateCameraModel = false;
        pCamera->bMoveCameraModel = false;
    }
    if (bFlyCamera)
    {
        updateFlyCamera();
        updateViewMatrix();
    }
    if (mEnableTransparency)
    {
        updatePolygonSorting();
    }

    for (unsigned int i = 0, size = mShaderVec.size();
         i < size; ++i)
    {
        renderShader(mShaderVec[i]);
    }
    {
        TRACE_TIME;
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, 0);
        mCurrentImage = "";
        mCurrentTextNumber = -1;
    }
}

void RenderScene::primitivePolygonSorting(RenderPrimitive* pPrimitive)
{
    Node* pNode = pPrimitive->getNode();
    glm::mat4 viewNatrix =
        pCamera->getModelViewMatrix(pCamera->getLook(),
                                    pNode->getGlobalMatrix(),
                                    mCameraMoveTime,
                                    mPreCameraTime);
    pPrimitive->polyonSorting(viewNatrix);
}

void RenderScene::updatePolygonSorting()
{
    bool viewChanged = false;
    if (mEnableRotation)
    {
        viewChanged = true;
    }
    glm::mat4 currentView = pCamera->getLook();
    for (unsigned int i = 0; i < 4; i++)
    {
        for (unsigned int j = 0; j < 4; j++)
        {
            if (currentView[i][j] != mLastModelView[i][j])
            {
                viewChanged = true;
                break;
            }
        }
    }
    if (!viewChanged)
    {
        return;
    }
    mLastModelView = currentView;
    unsigned int shaderSize = mShaderVec.size();
    for (unsigned int i = 0; i < shaderSize; ++i)
    {
        RenderShader* pRenderShader = mShaderVec[i];
        unsigned int primitiveSize = pRenderShader->getRenderPrimSize();
#if OPENMPENABLE
        int loopSize = primitiveSize / WAVESIZE;
        int remainder = primitiveSize - loopSize * WAVESIZE;
        for (int k = 0; k < loopSize; ++k)
        {
#pragma omp parallel for
            for (int j = 0; j < WAVESIZE; ++j)
            {
                primitivePolygonSorting
                    (pRenderShader->getRenderPrim(k * WAVESIZE + j));
            }
        }
        for (int k = 0; k < remainder; ++k)
        {
            primitivePolygonSorting
                (pRenderShader->getRenderPrim(loopSize * WAVESIZE + k));
        }
#else
        for (unsigned int j = 0; j < primitiveSize; ++j)
        {
            primitivePolygonSorting(pRenderShader->getRenderPrim(j));
        }
#endif
    }
}

void RenderScene::renderShader(RenderShader* pRenderShader)
{
    Technique* pTechnique = pRenderShader->getTechnique();

    if (!pTechnique->useTechnique())
    {
        return;
    }

    unsigned int progId = pTechnique->getProgramId();

    {
        TRACE_TIME;
        upLoadTechInfo(progId, pTechnique);
    }

    for (unsigned int i = 0, size = pRenderShader->getRenderPrimSize();
         i < size; ++i)
    {
        renderPrimitive(pRenderShader->getRenderPrim(i), progId);
    }
}

void RenderScene::completeRender()
{
#if ENABLE_FPS
    if (0 != pFPSCounter)
    {
        pFPSCounter->printFPS(&mCurrentViewport);
    }
#endif

    {
        TRACE_TIME;
        glViewport(mCurrentViewport.x, mCurrentViewport.y,
                   mCurrentViewport.width, mCurrentViewport.height);
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
        fbo.renderFboTexture();
    }
    return;
}

void RenderScene::releaseRender()
{
    mLoadJson.releaseFileName();
    fbo.releaseFbo();
    return;
}

CPhysicalCamera* RenderScene::getCamera()
{
    return pCamera;
}

Node* RenderScene::findNodeByName(Node* pPareNode,
                                  const std::string& nodeId)
{
    if (0 == pPareNode)
        return 0;
    if (pPareNode->getNodeName() == nodeId)
        return pPareNode;

    Node* pNode = 0;
    for (unsigned int i = 0, size = pPareNode->getChildNodeSize();
         i < size; ++i)
    {
        pNode = findNodeByName(pPareNode->getChildNode(i), nodeId);
        if (0 != pNode)
            break;
    }
    return pNode;
}

Node* RenderScene::findNodeByJoint(Node* pPareNode,
                                   const std::string& jointId)
{
    if (0 == pPareNode || !pPareNode->getJointFlag())
        return 0;
    if (pPareNode->getJointId() == jointId)
        return pPareNode;

    Node* pNode = 0;
    for (unsigned int i = 0, size = pPareNode->getChildNodeSize();
         i < size; ++i)
    {
        pNode = findNodeByJoint(pPareNode->getChildNode(i), jointId);
        if (0 != pNode)
            break;
    }
    return pNode;
}

int RenderScene::prepareRenderBitmap(glTFViewport* pViewport)
{
    int status = initSSAAFrameBuf(pViewport);
    if (LIBGLTF_SUCCESS != status)
    {
        return status;
    }
    startAnimation();
    glEnable(GL_DEPTH_TEST);
    return LIBGLTF_SUCCESS;
}

void RenderScene::renderBitmap(double time)
{
    mCurrentTime = time;
    realRender();
}

void gaussianFilter(unsigned char* corrupted, glTFViewport* pViewport)
{
    int templates[25] = { 1, 4,  7,  4,  1,
                          4, 16, 26, 16, 4,
                          7, 26, 41, 26, 7,
                          4, 16, 26, 16, 4,
                          1, 4,  7,  4,  1 };
    int desLineSize = (pViewport->width * 3 + 3) / 4 * 4;
    int size = desLineSize * pViewport->height;
    unsigned char* smooth = new unsigned char[size];
    memcpy(smooth, corrupted, size);
    for (int j = 2; j < pViewport->height - 2; j++)
    {
        for (int i = 2; i < pViewport->width - 2; i++)
        {
            int sum = 0;
            int index = 0;
            for ( int m = j - 2; m < j + 3; m++)
            {
                for (int n = i - 2; n < i + 3; n++)
                {
                    sum += corrupted[m * desLineSize + n * 3] *
                           templates[index++];
                }
            }
            sum /= 273;
            if (sum > 255)
                sum = 255;
            smooth[j * desLineSize + i * 3] = (unsigned char)sum;
        }
    }
    memcpy(corrupted, smooth, size);
    delete[] smooth;
    smooth = NULL;
}

void RenderScene::setBitZoom(unsigned char* Dstbuffer,
                             unsigned char* Srcbuffer,
                             glTFViewport* pViewport,
                             int bufferDepth)
{
#if DEFAULT_VIEW
    unsigned int width  = SSAA * DEFAULT_VIEW;
    unsigned int height = (SSAA * DEFAULT_VIEW *
                           pViewport->height / pViewport->width);
#else
    unsigned int width  = SSAA * pViewport->width;
    unsigned int height = SSAA * pViewport->height;
#endif //DEFAULT_VIEW
    int dstLineByte = pViewport->width * bufferDepth;
    int srcLineByte = width * bufferDepth;
    for (int i = 0; i < pViewport->height; i++)
    {
        float fy = (float)((i + 0.5) * SSAA - 0.5);
        int sy =(int)floor(fy);
        fy -= sy;
        sy = min(sy, (int)(height - 2));
        sy = max(0, sy);
        short cbufy[2];
        cbufy[0] = (short)((1.0f - fy) * 2048);
        cbufy[1] = 2048 - cbufy[0];
        for (int j = 0; j < pViewport->width; j++)
        {
            float fx = (float)((j + 0.5) * SSAA - 0.5);
            unsigned int sx = (int)floor(fx);
            fx -= sx;
            if (sx >= width - 1)
            {
                fx = 0;
                sx = width - 2;
            }
            short cbufx[2];
            cbufx[0] = (short)((1.0f - fx) * 2048);
            cbufx[1] = 2048 - cbufx[0];
            for (int k = 0; k < bufferDepth; k++)
            {
                *(Dstbuffer + i * dstLineByte + bufferDepth * j + k) = (
                    *(Srcbuffer + sy * srcLineByte + bufferDepth * sx + k)
                    * cbufx[0] * cbufy[0] +
                    *(Srcbuffer + (sy + 1) * srcLineByte + bufferDepth * sx + k)
                    * cbufx[0] * cbufy[1] +
                    *(Srcbuffer + sy * srcLineByte + bufferDepth * (sx + 1) + k)
                    * cbufx[1] * cbufy[0] +
                    *(Srcbuffer + (sy + 1) * srcLineByte + bufferDepth
                    *(sx + 1) + k) * cbufx[1] * cbufy[1]) >> 22;
            }
        }
    }
}


void RenderScene::completeRenderBitmap(glTFViewport* pViewport,
                                       unsigned char* buffer, GLenum format)
{
#if DEFAULT_VIEW
    unsigned int width  = SSAA * DEFAULT_VIEW;
    unsigned int height = (SSAA * DEFAULT_VIEW *
                           pViewport->height / pViewport->width);
#else
    unsigned int width  = SSAA * pViewport->width;
    unsigned int height = SSAA * pViewport->height;
#endif //DEFAULT_VIEW
    glDisable(GL_DEPTH_TEST);
    fbo.createAndBindBitMapFBO(width, height);
    glViewport(pViewport->x, pViewport->y, width, height);
    fbo.renderFboTexture();
    fbo.inverseBitMap(width, height);
    if (GL_FRAMEBUFFER_COMPLETE != glCheckFramebufferStatus(GL_FRAMEBUFFER))
    {
        return;
    }
    int bufferDepth;
    if (format == GL_RGB || format == GL_BGR)
        bufferDepth = 3;
    else if (format == GL_RGBA || format == GL_BGRA)
        bufferDepth = 4;
    else
        return;
    int imageSize = width * height * bufferDepth;

    unsigned char *pbuffer = new unsigned char[imageSize];
    glReadPixels(0, 0, width, height, format,
                 GL_UNSIGNED_BYTE, pbuffer);
    setBitZoom(buffer, pbuffer, pViewport, bufferDepth);
    if (pbuffer)
    {
        delete[] pbuffer;
        pbuffer = NULL;
    }
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    fbo.releaseBitMapFBO();
    fbo.releaseInverseBitMapFBO();
}

void RenderScene::getCameraPos(glm::vec3* Eye, glm::vec3* view,
                               glm::vec3* up)
{
    *Eye =  pCamera->vEye;
    *view = glm::vec3(pCamera->vView.x - 5,
                      pCamera->vView.y - 5.0f,
                      pCamera->vView.z - 5.0f);
    *up = pCamera->vUp;
}

glm::vec3* RenderScene::getModelCenterPos()
{
    return &pCamera->vModelCenterPos;
}

void RenderScene::EnableRotation()
{
    mEnableRotation = true;
}

void RenderScene::DisableRotation()
{
    mEnableRotation = false;
}

void RenderScene::EnableTransparency()
{
    mEnableTransparency = true;
}

void RenderScene::DisableTransparency()
{
    mEnableTransparency = false;
}

double RenderScene::getModelSize()
{
    return pCamera->flength;
}

void RenderScene::startAnimation()
{
    this->mAnimationPlay = true;
    mCurrentTime = 0;
    mUpdateTimeOut = 0;
    bAnimation = true;
}

void RenderScene::stopAnimation()
{
    if (pScene->getSkinSize() != 0)
    {
        this->mAnimationPlay = false;
        bAnimation = false;
    }
}

void RenderScene::stopAerialView()
{
    pCamera->bAerialView = false;
    roCameraX = glm::mat4(1.0);
    roCameraY = glm::mat4(1.0);
    pCamera->mTrackBallY= glm::mat4(1.0);
    pCamera->mTrackBallX= glm::mat4(1.0);
    pCamera->AerialViewY = 0.0;
    trackball(pCamera->RotateCurquatY, 0.0f, 0.0f, 0.0f, 0.0f);
    trackball(pCamera->RotateCurquatX, 0.0f, 0.0f, 0.0f, 0.0f);
    bAerialView = false;
    if (IsAerialMode)
    {
        pScene->setUseCameraInJson(true);

        cCamera->getCameraNode()->setLocalMatrix(initCameraLocalMatrix);
        cCamera->getCameraNode()->setGlobalMatrix(initCameraGlobalMatrix);
        pCamera->bRotateCameraModel = true;
    }
    updateViewMatrix();
}

void RenderScene::startAerialView()
{
    pCamera->bAerialView = true;
    roCameraX = glm::mat4(1.0);
    roCameraY = glm::mat4(1.0);
    pCamera->mTrackBallY= glm::mat4(1.0);
    pCamera->mTrackBallX= glm::mat4(1.0);
    trackball(pCamera->curquatY, 0.0f, 0.0f, 0.0f, 0.0f);
    trackball(pCamera->curquatX, 0.0f, 0.0f, 0.0f, 0.0f);
    bAerialView = true;
    if (pScene->getUseCameraInJson())
    {
        IsAerialMode = true;
        pScene->setUseCameraInJson(false);
    }
    pCamera->vView = tagView;
    pCamera->mLook = glm::lookAt(pCamera->vEye, pCamera->vView, pCamera->vUp);
    updateViewMatrix();
}

void RenderScene::setAnimTime(double time)
{
    this->mCurrentTime = time;
    this->mUpdateTimeOut = time;
}

double RenderScene::getAnimTime()
{
    errno = 0;
    double time = fmod(this->mCurrentTime, this->mDuration);
    return errno == EDOM ? 0.0 : time;
}

void RenderScene::setAnimLoop(int loop)
{
    if (0 != loop)
    {
        this->mAnimationLoop = true;
    }
    else
    {
        this->mAnimationLoop = false;
    }
}

int  RenderScene::getAnimLoop()
{
    return this->mAnimationLoop ? 1 : 0;
}

double RenderScene::getAnimDuration()
{
    return this->mDuration;
}

int  RenderScene::isAnimPlay()
{
    return this->mAnimationPlay ? 1 : 0;
}

void RenderScene::resumeAnim()
{
    bAnimation = true;
    this->mAnimationPlay = true;
}

void RenderScene::setTimeForAnim()
{
    double timeNow = libgltf::time::getCurrentTime();
    if (mAnimationPlay)
    {
        if (bIsTimeAvailable)
        {
            mCurrentTime += libgltf::time::diffTime(timeNow,mLastPlaying);
        }
        else
        {
            bIsTimeAvailable = true;
        }
        mLastPlaying = timeNow;
    }
    else
    {
        bIsTimeAvailable = false;
    }
    if (!mAnimationLoop && mCurrentTime > mDuration)
    {
        stopAnimation();
        setAnimTime(0.0);
    }
}

void RenderScene::getCameraInfo(cameraInfo* pCameraInfo)
{
    glm::mat4& viewMatrix = pCamera->getLook();
    for (int i = 0; i< 4; ++i)
    {
        for (int j = 0; j < 4; ++j)
        {
            pCameraInfo->mat4x4[i * 4 + j] = viewMatrix[i][j];
        }
    }
}

void RenderScene::renderFlyCamera(glm::vec3 glPosInfo, double time)
{
    glm::mat4 newViewMatrix = glm::lookAt(glPosInfo,glm::vec3(0,0,0),pCamera->vUp);
    glm::mat4 oldViewMatrix = pCamera->getLook();
    flyInfo = newViewMatrix - oldViewMatrix;
    if (!time)
    {
        pCamera->mLook = newViewMatrix;
        return;
    }
    bFlyCamera = true;
    flyTime = time * 1000 * 1000;
    flyInfo = flyInfo / (float)(time*1000*1000);
}

/* vim:set shiftwidth=4 softtabstop=4 expandtab: */
